import json
from copy import deepcopy
from dataclasses import dataclass
from functools import partial
from pprint import pprint

from typing import List, Self, Any, Callable
from datetime import date, datetime

from oldaplib.src.cachesingleton import CacheSingletonRedis
from oldaplib.src.enums.adminpermissions import AdminPermission
from oldaplib.src.enums.projectattr import ProjectAttr
from oldaplib.src.helpers.context import Context
from oldaplib.src.enums.action import Action
from oldaplib.src.dtypes.namespaceiri import NamespaceIRI
from oldaplib.src.helpers.irincname import IriOrNCName
from oldaplib.src.helpers.observable_dict import ObservableDict
from oldaplib.src.helpers.serializer import serializer
from oldaplib.src.helpers.tools import lprint
from oldaplib.src.xsd.iri import Iri
from oldaplib.src.xsd.xsd_anyuri import Xsd_anyURI
from oldaplib.src.xsd.xsd_qname import Xsd_QName
from oldaplib.src.xsd.xsd_ncname import Xsd_NCName
from oldaplib.src.xsd.xsd_date import Xsd_date
from oldaplib.src.xsd.xsd_datetime import Xsd_dateTime
from oldaplib.src.helpers.langstring import LangString
from oldaplib.src.helpers.oldaperror import OldapError, OldapErrorAlreadyExists, OldapErrorNoPermission, \
    OldapErrorUpdateFailed, OldapErrorNotFound, OldapErrorInconsistency
from oldaplib.src.helpers.query_processor import QueryProcessor
from oldaplib.src.iconnection import IConnection
from oldaplib.src.model import Model
from oldaplib.src.helpers.attributechange import AttributeChange
from oldaplib.src.xsd.xsd_string import Xsd_string

@serializer
@dataclass(frozen=True)
class ProjectSearchResult:
    projectIri: IriOrNCName
    projectShortName: Xsd_NCName

    def _as_dict(self):
        return {'projectIri': self.projectIri, 'projectShortName': self.projectShortName}


#@strict
@serializer
class Project(Model):
    """
    # Project

    This class implements the Project model. A Project is a distinct research space within Oldap
    framework that offers dedicated space for data, its own data modeling and access control. A project
    needs the following metadata:

    - `projectIri: Iri | None` [optional]:
      The IRI that uniquely identifies this project. This can be an
      [Iri](/python_docstrings/datatypes#oldaplib.src.helpers.datatypes.Iri) instance or a string with either an Iri or Qname
      [QName](/python_docstrings/datatypes#oldaplib.src.helpers.datatypes.QName). If the `projectIri` is ommitted,
      a random IRI based on the URN semantics will be generated.
    - `projectShortName: Xsd_NCName | str` [mandatory]:
      The short name of the project that must be a NCName or a string
    - `label: LangString | None` [optional]:
      A multilingual string with a human-readable label for the project (`rdfs:label`)
    - `comment: LangString | None` [optional]:
      A multilingual description of the project (`rdfs:comment`)
    - `namespaceIri: NamespaceIri` [mandatory]:
       The namespace that the project uses for its data and data model. Must be
       a [NamespaceIRI](/python_docstrings/datatypes#oldaplib.src.helpers.datatypes.NamespaceIRI).
    - `projectStart: Xsd_date | date | None` [optional]:
      The start date of the project.  Must be a Python `date` type. If not set,
       the current date will be used.
    - `projectEnd: Xsd_date | None` [optional]:
      The optional end date of the project. Must be a Xsd_date or a Python `date` type.

    The class provides the following methods:

    - [Project(...)](/python_docstrings/project#oldaplib.src.project.Project.__init__)): Constructor
    - [read(...)](/python_docstrings/project#oldaplib.src.project.Project.read):
      Read project form triplestore and return a Project-instance
    - [search(...)](/python_docstrings/project#oldaplib.src.project.Project.search):
      Search a Project in the triplestore
    - [create(...)](/python_docstrings/project#oldaplib.src.project.Project.create): Create a new project in the triplestore
    - [update(...)](/python_docstrings/project#oldaplib.src.project.Project.update):
      Write changes to triplestore
    - [delete(...)](/python_docstrings/project#oldaplib.src.project.Project.update):
      Delete a project in the triplestore

    The class provides the following properties:

    - `projectIri` [read only]: The project IRI
    - `projectShortName` [read/write]: The project short name. Must be a NCName
    - `projectLabel` [read/write]: The projects label as multilingual LangString
    - `projectComment` [read/write]: The projects comment/description as multilingual
      [LangString](/python_docstrings/langstring)
    - `namespaceIri` [read only]: The namespace that the project uses for its data and data model. Must be
      a [NamespaceIRI](/python_docstrings/datatypes#oldaplib.src.helpers.datatypes.NamespaceIRI).
    - `projectStart` [read/write]: The start date of the project. Must be a Python `date` type
    - `projectEnd` [read/write]: The end date of the project. Must be a Python `date` type
    - `creator` [read only]: The creator of the project. Must be a
      [AnyIRI](/python_docstrings/datatypes#oldaplib.src.helpers.datatypes.AnyIRI) or
      [QName](/python_docstrings/datatypes#oldaplib.src.helpers.datatypes.AnyIRI)
    - `created` [read only]: The creation date of the project. Must be a Python `date` type
    - `contributor` [read only]: The person which made the last changes to the project data. Must be a
      [AnyIRI](/python_docstrings/datatypes#oldaplib.src.helpers.datatypes.AnyIRI) or
      [QName](/python_docstrings/datatypes#oldaplib.src.helpers.datatypes.AnyIRI)
    - `modified` [read only]: The modification date of the project. Must be a Python `date` type

    """

    #_attributes: dict[ProjectAttr, ProjectAttrTypes]
    #__changeset: dict[ProjectAttr, ProjectAttrChange]

    __slots__ = ('projectIri', 'projectShortName', 'label', 'comment', 'namespaceIri', 'projectStart', 'projectEnd')

    def __init__(self, *,
                 con: IConnection,
                 creator: Iri | str | None = None,
                 created: Xsd_dateTime | datetime | str | None = None,
                 contributor: Iri | str | None = None,
                 modified: Xsd_dateTime | datetime | str | None = None,
                 validate: bool = False,
                 **kwargs):
        """
        Constructs a new Project
        :param con: [Connection](/python_docstrings/iconnection) instance
        :param creator: Creator of the project  [DO NOT SET!]
        :type creator: Xsd_anyURI | None
        :param created: Date the project was created  [DO NOT SET!]
        :type created: datetime | None
        :param contributor: person that made the last change  [DO NOT SET!]
        :type contributor: Xsd_anyURI
        :param modified: Last date the project was modified  [DO NOT SET!]
        :type modified: date | None
        :param projectIri: IRI to be used for the project. If no projectIRI is provied, the constrctor
         will create an arbitrary IRI based on thr URN scheme and a UUID. [optional].
        :type projectIri: Iri
        :param projectShortName: A short name for the project. Is used as prefix for named graphs that
           are being used for the project. [mandatory]
        :type projectShortName: NCname (strings are accepted only if conform to NCName syntax)
        :param namespaceIri: The namespace to be used for the projects data [mandatory]
        :type namespaceIri: NamespaceIRI
        :param label: Human-readable name for the project (multi-language) (`rdfs:label`) [optional]
        :type label: LangString
        :param comment: Description of the project (multi-language) (`rdfs:comment`)
        :type comment: LangString
        :param projectStart: Start date of the project
        :type projectStart: Python date
        :param projectEnd: End date of the project [Optional]
        :type projectEnd: Python date
        :raises OldapErrorValue: Invalid parameter supplied
        """
        super().__init__(connection=con,
                         created=created,
                         creator=creator,
                         modified=modified,
                         contributor=contributor,
                         validate=validate)

        self.set_attributes(kwargs, ProjectAttr)
        #
        # Consistency checks
        #
        if not self._attributes.get(ProjectAttr.PROJECT_IRI):
            self._attributes[ProjectAttr.PROJECT_IRI] = Iri()
        if not self._attributes.get(ProjectAttr.PROJECT_START):
            self._attributes[ProjectAttr.PROJECT_START] = Xsd_date()
        if self._attributes.get(ProjectAttr.PROJECT_END):
            self.check_consistency(ProjectAttr.PROJECT_END, self._attributes[ProjectAttr.PROJECT_END])
        if self._attributes.get(ProjectAttr.PROJECT_START):
            self.check_consistency(ProjectAttr.PROJECT_START, self._attributes[ProjectAttr.PROJECT_START])
        #
        # create all the attributes of the class according to the ProjectFields dfinition
        #
        for attr in ProjectAttr:
            setattr(Project, attr.value.fragment, property(
                partial(Project._get_value, attr=attr),
                partial(Project._set_value, attr=attr),
                partial(Project._del_value, attr=attr)))
        self._changeset = {}

    def update_notifier(self):
        for attr, value in self._attributes.items():
            if getattr(value, 'set_notifier', None) is not None:
                value.set_notifier(self.notifier, attr)

    def _as_dict(self):
        return {x.fragment: y for x, y in self._attributes.items()} | super()._as_dict()

    def __deepcopy__(self, memo: dict[Any, Any]) -> Self:
        if id(self) in memo:
            return memo[id(self)]
        cls = self.__class__
        instance = cls.__new__(cls)
        memo[id(self)] = instance
        Model.__init__(instance,
                       connection=deepcopy(self._con, memo),
                       creator=deepcopy(self._creator, memo),
                       created=deepcopy(self._created, memo),
                       contributor=deepcopy(self._contributor, memo),
                       modified=deepcopy(self._modified, memo))
        # Copy internals of Model:
        instance._attributes = deepcopy(self._attributes, memo)
        instance._changset = deepcopy(self._changeset, memo)
        return instance

    def __eq__(self, other: Self) -> bool:
        if not isinstance(other, Project):
            return False
        return self._as_dict() == other._as_dict()

    def check_consistency(self, attr: ProjectAttr, value: Any) -> None:
        if attr == ProjectAttr.PROJECT_END:
            if self._attributes.get(ProjectAttr.PROJECT_START) and value < self._attributes[ProjectAttr.PROJECT_START]:
                raise OldapErrorInconsistency(f'Project start date {self._attributes[ProjectAttr.PROJECT_START]} is after project end date {value}.')

        if attr == ProjectAttr.PROJECT_START:
            if self._attributes.get(ProjectAttr.PROJECT_END) and value > self._attributes[ProjectAttr.PROJECT_END]:
                raise OldapErrorInconsistency(f'Project start date {value} is after project end date {self._attributes[ProjectAttr.PROJECT_END]}.')

    def check_for_permissions(self) -> (bool, str):
        actor = self._con.userdata
        sysperms = actor.inProject.get(Iri('oldap:SystemProject'))
        if sysperms and AdminPermission.ADMIN_OLDAP in sysperms:
            return True, "OK"
        else:
            return False, "No permission to create/change a project."

    def notifier(self, attr: ProjectAttr) -> None:
        """
        This method is called when a field is being changed.
        :param fieldname: Fieldname of the field being modified
        :return: None
        """
        self._changeset[attr] = AttributeChange(self._attributes[attr], Action.MODIFY)



    @classmethod
    def read(cls,
             con: IConnection,
             projectIri_SName: IriOrNCName | Iri | Xsd_NCName | str,
             ignore_cache: bool = False) -> Self:
        """
        Reads the project from the triplestore and returns an instance of the project.

        This method queries the triplestore based on a provided IRI or shortname of the project,
        optionally utilizing a cache for optimization. If the project is found, it fetches and
        processes the project's data including metadata and creates an instance of the project.
        If not found, an appropriate error is raised.

        :param con: A valid connection to the triplestore.
        :type con: IConnection
        :param projectIri_SName: The IRI or shortname of the project to be read.
        :type projectIri_SName: IriOrNCName | Iri | Xsd_NCName | str
        :param ignore_cache: Whether to bypass the cache and query the triplestore directly.
        :type ignore_cache: bool
        :return: An instance of the project with the fetched data.
        :rtype: Self
        :raises OldapErrorNotFound: If the project with the specified IRI or shortname is not found.
        :raises OldapError: If other errors or problems occur during the process.
        """
        context = Context(name=con.context_name)
        query = context.sparql_context

        if not isinstance(projectIri_SName, IriOrNCName):
            projectIri_SName = IriOrNCName(projectIri_SName, validate=True)
        shortname, projectIri = projectIri_SName.value()
        # projectIri: Iri | None = None
        # shortname: Xsd_NCName | None = None
        # if isinstance(projectIri_SName, Iri):
        #     projectIri = projectIri_SName
        # elif isinstance(projectIri_SName, Xsd_NCName):
        #     shortname = Xsd_NCName(projectIri_SName)
        # else:
        #     if ':' in str(projectIri_SName):  # must be IRI or QName
        #         projectIri = Iri(projectIri_SName)
        #     else:
        #         shortname = Xsd_NCName(projectIri_SName)
        cache = CacheSingletonRedis()
        if projectIri is not None:
            if not ignore_cache:
                tmp = cache.get(projectIri, connection=con)
                if tmp is not None:
                    tmp.update_notifier()
                    return tmp
            query += f"""
                SELECT ?prop ?val
                FROM NAMED oldap:admin
                WHERE {{
                    GRAPH oldap:admin {{
                        {projectIri.toRdf} ?prop ?val
                    }}
                }}
            """
        elif shortname is not None:
            if not ignore_cache:
                tmp = cache.get(shortname, connection=con)
                if tmp is not None:
                    tmp._con = con
                    return tmp
            query += f"""
                SELECT ?proj ?prop ?val ?prefix ?iri
                WHERE {{
                    GRAPH oldap:admin {{
                        ?proj a oldap:Project .
                        ?proj oldap:projectShortName {shortname.toRdf} .
                        ?proj ?prop ?val .
                        OPTIONAL {{
                            ?val oldap:prefix ?prefix .
                            ?val oldap:fullIri ?iri
                        }}
                    }}
                }}
            """
        jsonobj = con.query(query)
        res = QueryProcessor(context, jsonobj)
        if len(res) == 0:
            raise OldapErrorNotFound(f'Project with IRI/shortname "{projectIri_SName}" not found.')
        creator: Iri | None = None
        created: Xsd_dateTime | None = None
        contributor: Iri | None = None
        modified: Xsd_dateTime | None = None
        projectShortName = None
        namespaceIri = None
        label = LangString()
        comment = LangString()
        projectStart = None
        projectEnd = None
        for r in res:
            if projectIri is None:
                projectIri = r['proj']
            match str(r.get('prop')):
                case 'dcterms:creator':
                    creator = r['val']
                case 'dcterms:created':
                    created = r['val']
                case 'dcterms:contributor':
                    contributor = r['val']
                case 'dcterms:modified':
                    modified = r['val']
                case 'oldap:namespaceIri':
                    namespaceIri = NamespaceIRI(r['val'])
                case 'oldap:projectShortName':
                    projectShortName = r['val']
                case 'rdfs:label':
                    label.add(str(r['val']))
                case 'rdfs:comment':
                    comment.add(str(r['val']))
                case 'oldap:projectStart':
                    projectStart = r['val']
                case 'oldap:projectEnd':
                    projectEnd = r['val']
        if label:
            label.clear_changeset()
            label.set_notifier(cls.notifier, Xsd_QName(ProjectAttr.LABEL.value))
        if comment:
            comment.clear_changeset()
            comment.set_notifier(cls.notifier, Xsd_QName(ProjectAttr.COMMENT.value))
        context[projectShortName] = namespaceIri
        instance = cls(con=con,
                       creator=creator,
                       created=created,
                       contributor=contributor,
                       modified=modified,
                       projectIri=projectIri,
                       projectShortName=projectShortName,
                       label=label,
                       namespaceIri=namespaceIri,
                       comment=comment,
                       projectStart=projectStart,
                       projectEnd=projectEnd)
        cache = CacheSingletonRedis()
        cache.set(instance.projectIri, instance, instance.projectShortName)
        return instance

    @staticmethod
    def search(con: IConnection,
               label: Xsd_string | str | None = None,
               comment: Xsd_string | str | None = None) -> list[ProjectSearchResult]:
        """
        Search for projects based on the provided label and/or comment. If no label or comment is provided, all
        existing projects are returned. When both label and comment are specified, the search combines their
        conditions using logical AND.

        :param con: Valid connection object representing the current context
        :type con: IConnection
        :param label: String to search within project labels. The label search is performed by checking
            if a project's label contains the specified string.
        :type label: Xsd_string or str or None
        :param comment: String to search within project comments. The comment search is performed by
            checking if a project's comment contains the specified string.
        :type comment: Xsd_string or str or None
        :return: List of project search results matching the specified criteria.
        :rtype: list[ProjectSearchResult]
        :raises OldapErrorNotFound: Raised if the project does not exist in the repository or cannot
            be located based on the search criteria.
        """
        label = Xsd_string(label, validate=True)
        comment = Xsd_string(comment, validate=True)
        context = Context(name=con.context_name)
        sparql = context.sparql_context
        sparql += 'SELECT DISTINCT ?project ?shortname\n'
        sparql += 'FROM oldap:onto\n'
        sparql += 'FROM shared:onto\n'
        sparql += 'FROM NAMED oldap:admin\n'
        sparql += 'WHERE {\n'
        sparql += '    GRAPH oldap:admin {\n'
        sparql += '        ?project a oldap:Project .\n'
        sparql += '        ?project oldap:projectShortName ?shortname .\n'
        if label:
            sparql += '        ?project rdfs:label ?label .\n'
            sparql += '    }\n'
            sparql += f'    FILTER(CONTAINS(STR(?label), "{Xsd_string.escaping(label.value)}"))\n'
        if comment:
            sparql += '        ?project rdfs:comment ?comment .\n'
            sparql += '    }\n'
            sparql += f'    FILTER(CONTAINS(STR(?comment), "{Xsd_string.escaping(comment.value)}"))\n'
        if not label and not comment:
            sparql += '    }\n'
        sparql += '}\n'
        jsonobj = con.query(sparql)
        res = QueryProcessor(context, jsonobj)
        projects: list[tuple(Iri, Xsd_NCName)] = []
        if len(res) > 0:
            for r in res:
                projects.append(ProjectSearchResult(r['project'], r['shortname']))
        return projects


    def trig_to_str(self, created: Xsd_dateTime, modified: Xsd_dateTime,indent: int = 0, indent_inc: int = 4):
        blank = ''
        sparql = ''
        sparql += f'\n{blank:{indent * indent_inc}}{self.projectIri.toRdf} a oldap:Project'
        sparql += f' ;\n{blank:{(indent + 1) * indent_inc}}dcterms:creator {self._con.userIri.toRdf}'
        sparql += f' ;\n{blank:{(indent + 1) * indent_inc}}dcterms:created {created.toRdf}'
        sparql += f' ;\n{blank:{(indent + 1) * indent_inc}}dcterms:contributor {self._con.userIri.toRdf}'
        sparql += f' ;\n{blank:{(indent + 1) * indent_inc}}dcterms:modified {modified.toRdf}'
        for attr, value in self._attributes.items():
            if not value:
                continue
            sparql += f' ;\n{blank:{(indent + 1) * indent_inc}}{attr.value.toRdf} {value.toRdf}'
        return sparql

    def create(self, indent: int = 0, indent_inc: int = 4) -> None:
        """
        Creates a new project in the triple store. The function starts by
        checking if the connection is available, verifies if the user
        has the required permissions, and then proceeds to execute
        SPARQL queries to ensure the project does not already exist.
        If the project doesn't exist, it creates the project by
        constructing and executing SPARQL INSERT queries. It also
        updates the project metadata, including timestamps and user
        information.

        :param indent: The start indent level for the generated SPARQL
            query, primarily used for debugging.
        :type indent: int
        :param indent_inc: The increment for indentation level, which
            affects how the SPARQL query is formatted.
        :type indent_inc: int
        :return: None
        :raises OldapErrorAlreadyExists: If a project already exists with
            the same project IRI.
        :raises OldapError: If any other errors occur during execution.
        """
        if self._con is None:
            raise OldapError("Cannot create: no connection")
        #
        # First we check if the logged-in user ("actor") has the permission to create a user for
        # the given project!
        #
        result, message = self.check_for_permissions()
        if not result:
            raise OldapErrorNoPermission(message)

        timestamp = Xsd_dateTime.now()
        #indent: int = 0
        #indent_inc: int = 4

        context = Context(name=self._con.context_name)

        #
        # SPARQL to check if project with the given projectShortName already exists
        #
        sparql0 = context.sparql_context
        sparql0 += f"""
        ASK {{
            GRAPH oldap:admin {{
    	        ?p a oldap:Project ;
    		        oldap:projectShortName ?sname .
    	        FILTER(?sname = {self.projectShortName.toRdf})
	        }}
        }}
        """

        #
        # SPARQL to check if the given namespaceIri is already used by another project
        #
        sparql1a = context.sparql_context
        sparql1a += f"""
        ASK {{
            GRAPH oldap:admin {{
                ?project oldap:namespaceIri {self.namespaceIri.toRdf} .
            }}
        }}
        """

        #
        # SPARQL to check if project with the given projectIri already exists
        #
        sparql1b = context.sparql_context
        sparql1b += f"""
        ASK {{
            GRAPH oldap:admin {{
                {self.projectIri.toRdf} ?p ?o .
            }}
        }}
        """

        blank = ''
        sparql2 = context.sparql_context
        sparql2 += f'{blank:{indent * indent_inc}}INSERT DATA {{'
        sparql2 += f'\n{blank:{(indent + 1) * indent_inc}}GRAPH oldap:admin {{'
        sparql2 += self.trig_to_str(created=timestamp, modified=timestamp, indent=indent + 2, indent_inc=indent_inc)
        # sparql2 += f'\n{blank:{(indent + 2) * indent_inc}}{self.projectIri.toRdf} a oldap:Project'
        # sparql2 += f' ;\n{blank:{(indent + 3) * indent_inc}}dcterms:creator {self._con.userIri.toRdf}'
        # sparql2 += f' ;\n{blank:{(indent + 3) * indent_inc}}dcterms:created {timestamp.toRdf}'
        # sparql2 += f' ;\n{blank:{(indent + 3) * indent_inc}}dcterms:contributor {self._con.userIri.toRdf}'
        # sparql2 += f' ;\n{blank:{(indent + 3) * indent_inc}}dcterms:modified {timestamp.toRdf}'
        # for attr, value in self._attributes.items():
        #     if not value:
        #         continue
        #     sparql2 += f' ;\n{blank:{(indent + 3) * indent_inc}}{attr.value.toRdf} {value.toRdf}'
        sparql2 += f' .\n{blank:{(indent + 1) * indent_inc}}}}\n'
        sparql2 += f'{blank:{indent * indent_inc}}}}\n'

        self._con.transaction_start()
        result = self.safe_query(sparql0)
        if result['boolean']:
            self._con.transaction_abort()
            raise OldapErrorAlreadyExists(f'A Project with a shortname "{self.projectShortName}" already exists.')

        result = self.safe_query(sparql1a)
        if result['boolean']:
            self._con.transaction_abort()
            raise OldapErrorAlreadyExists(f'A Project with a namespace IRI "{self.namespaceIri}" already exists.')

        result = self.safe_query(sparql1b)
        if result['boolean']:
            self._con.transaction_abort()
            raise OldapErrorAlreadyExists(f'A Project with a project IRI "{self.projectIri}" already exists.')

        self.safe_update(sparql2)
        self.safe_commit()
        self._created = timestamp
        self._creator = self._con.userIri
        self._modified = timestamp
        self._contributor = self._con.userIri
        context[self._attributes[ProjectAttr.PROJECT_SHORTNAME]] = self._attributes[ProjectAttr.NAMESPACE_IRI]

        cache = CacheSingletonRedis()
        cache.set(self.projectIri, self, self.projectShortName)

    def update(self, indent: int = 0, indent_inc: int = 4) -> None:
        """
        Updates all modifications made to the project instance and writes them to the triple store.

        The function ensures changes contained within the project instance are written to the triple store
        by serializing them into SPARQL queries and executing them. Permissions are checked before
        execution, and transactions are used to ensure atomicity. If the update fails, appropriate
        exceptions are raised, and operations are rolled back.

        :param indent: Starting indentation level used for SPARQL query formatting.
        :type indent: int
        :param indent_inc: Increment of indentation for formatting SPARQL queries.
        :type indent_inc: int
        :return: None
        :raises OldapErrorNoPermission: Raised if the user does not have permissions for the operation.
        :raises OldapErrorUpdateFailed: Raised if the update fails due to timestamp mismatch or other inconsistencies.
        :raises OldapError: Raised for other internal errors.
        """

        def dict_diff(a: dict, b: dict) -> dict:
            """
            Compare two dicts and return added, removed, and changed key–values.
            """
            a_keys = set(a) if a else set()
            b_keys = set(b) if b else set()
            shared = a_keys & b_keys

            return {
                'added': {k: b[k] for k in b_keys - a_keys},
                'removed': {k: a[k] for k in a_keys - b_keys},
                'changed': {k: (a[k], b[k]) for k in shared if a[k] != b[k]},
                'same': {k: a[k] for k in shared if a[k] == b[k]},
            }

        result, message = self.check_for_permissions()
        if not result:
            raise OldapErrorNoPermission(message)

        timestamp = Xsd_dateTime.now()
        context = Context(name=self._con.context_name)
        blank = ''
        sparql_list = []
        for field, change in self._changeset.items():
            if field == ProjectAttr.LABEL or field == ProjectAttr.COMMENT:
                if change.action == Action.MODIFY:
                    sparql_list.extend(self._attributes[field].update(graph=Xsd_QName('oldap:admin'),
                                                                      subject=self.projectIri,
                                                                      field=Xsd_QName(field.value)))
                if change.action == Action.DELETE or change.action == Action.REPLACE:
                    sparql = self._changeset[field].old_value.delete(graph=Xsd_QName('oldap:admin'),
                                                                      subject=self.projectIri,
                                                                      field=Xsd_QName(field.value))
                    sparql_list.append(sparql)
                if change.action == Action.CREATE or change.action == Action.REPLACE:
                    sparql = self._attributes[field].create(graph=Xsd_QName('oldap:admin'),
                                                            subject=self.projectIri,
                                                            field=Xsd_QName(field.value))
                    sparql_list.append(sparql)
                continue

            sparql = f'{blank:{indent * indent_inc}}# Project field "{field.value}" with action "{change.action.value}"\n'
            if change.action != Action.CREATE:
                sparql += f'{blank:{indent * indent_inc}}DELETE {{\n'
                sparql += f'{blank:{(indent + 1) * indent_inc}}GRAPH oldap:admin {{\n'
                sparql += f'{blank:{(indent + 2) * indent_inc}}?project {field.value} {change.old_value.toRdf} .\n'
                sparql += f'{blank:{(indent + 1)* indent_inc}}}}\n'
                sparql += f'{blank:{indent * indent_inc}}}}\n'
            if change.action != Action.DELETE:
                sparql += f'{blank:{indent * indent_inc}}INSERT {{\n'
                sparql += f'{blank:{(indent + 1) * indent_inc}}GRAPH oldap:admin {{\n'
                sparql += f'{blank:{(indent + 2) * indent_inc}}?project {field.value} {self._attributes[field].toRdf} .\n'
                sparql += f'{blank:{(indent + 1) * indent_inc}}}}\n'
                sparql += f'{blank:{indent * indent_inc}}}}\n'
            sparql += f'{blank:{indent * indent_inc}}WHERE {{\n'
            sparql += f'{blank:{(indent + 1) * indent_inc}}BIND({self.projectIri.toRdf} as ?project)\n'
            if change.action != Action.CREATE:
                sparql += f'{blank:{(indent + 1) * indent_inc}}GRAPH oldap:admin {{\n'
                sparql += f'{blank:{(indent + 2) * indent_inc}}?project {field.value} {change.old_value.toRdf} .\n'
                sparql += f'{blank:{(indent + 1) * indent_inc}}}}\n'
            sparql += f'{blank:{indent * indent_inc}}}}'
            sparql_list.append(sparql)
        sparql = context.sparql_context
        sparql += " ;\n".join(sparql_list)

        self._con.transaction_start()
        try:
            self._con.transaction_update(sparql)
            self.set_modified_by_iri(Xsd_QName('oldap:admin'), self.projectIri, self.modified, timestamp)
            modtime = self.get_modified_by_iri(Xsd_QName('oldap:admin'), self.projectIri)
        except OldapError:
            print(sparql)
            self._con.transaction_abort()
            raise
        if timestamp != modtime:
            self._con.transaction_abort()
            raise OldapErrorUpdateFailed("Update failed! Timestamp does not match")
        self.safe_commit()
        self._modified = timestamp
        self._contributor = self._con.userIri
        self.clear_changeset()
        cache = CacheSingletonRedis()
        cache.set(self.projectIri, self, self.projectShortName)

    def delete(self) -> None:
        """
        Delete the specified user from the triplestore. The function ensures that the
        user has the necessary permissions before proceeding with the deletion process.
        Additionally, it cleans up related cache entries.

        :raises OldapErrorNoPermission: Raised if the user lacks the necessary
            permissions to perform the operation.
        :raises OldapError: Raised in case of generic internal errors.
        :return: None
        """
        result, message = self.check_for_permissions()
        if not result:
            raise OldapErrorNoPermission(message)

        #
        # TODO: Check if project has any datamodel and/or data. Decline the deletion if this is the case
        #
        context = Context(name=self._con.context_name)
        sparql = context.sparql_context
        sparql += f"""
        DELETE WHERE {{
            GRAPH oldap:admin {{
                {self.projectIri.toRdf} a oldap:Project .
                {self.projectIri.toRdf} ?prop ?val .
            }}
        }} 
        """
        self._con.update_query(sparql)
        cache = CacheSingletonRedis()
        cache.delete(self.projectIri)
        cache.delete(self.projectShortName)

    @staticmethod
    def get_shortname_from_iri(con: IConnection, iri: Iri) -> Xsd_NCName:
        """
        Extracts the project short name from the given IRI using a SPARQL query executed
        in the provided connection. The method communicates with the server to fetch
        the short name associated with the specified IRI. If the IRI does not resolve
        to exactly one short name, an exception is raised.

        :param con: The connection object used to interact with the data source.
        :type con: IConnection
        :param iri: An instance of Iri or a value representing the IRI to query.
        :type iri: Iri
        :return: The project short name associated with the provided IRI.
        :rtype: Xsd_NCName
        :raises OldapError: If the connection object is not an instance of IConnection.
        :raises OldapErrorNotFound: If no short name is found or if multiple results are
            returned for the specified IRI.
        """
        if not isinstance(con, IConnection):
            raise OldapError("Connection must be an instance of IConnection")
        if not isinstance(iri, Iri):
            iri = Iri(iri, validate=True)
        context = Context(name=con.context_name)
        sparql = context.sparql_context
        sparql += 'SELECT ?shortname\n'
        sparql += 'FROM oldap:onto\n'
        sparql += 'FROM shared:onto\n'
        sparql += 'FROM NAMED oldap:admin\n'
        sparql += 'WHERE {\n'
        sparql += '    GRAPH oldap:admin {\n'
        sparql += f'        {iri.toRdf} oldap:projectShortName ?shortname .\n'
        sparql += '    }\n'
        sparql += '}\n'
        jsonobj = con.query(sparql)
        res = QueryProcessor(context, jsonobj)
        if len(res) != 1:
            raise OldapErrorNotFound(f"No project shortname found for {iri}")
        return res[0]['shortname']





import json
from enum import Enum, unique, EnumMeta

from oldaplib.src.helpers.serializer import serializer

class _ValidateEnumMeta(EnumMeta):
    def __call__(cls, value, *args, **kwargs):
        validate = kwargs.pop("validate", True)  # accept validate kwarg
        try:
            return super().__call__(value, *args, **kwargs)
        except ValueError:
            if validate:
                raise
            return None  # or: return a default, or keep ValueError

@unique
@serializer
class Editor(Enum, metaclass=_ValidateEnumMeta):
    AUTO_COMPLETE = "dash:AutoCompleteEditor"
    BLANK_NODE = "dash:BlankNodeEditor"
    BOOLEAN_SELECT = "dash:BooleanSelectEditor"
    DATE_PICKER = "dash:DatePickerEditor"
    DATE_TIME_PICKER = "dash:DateTimePickerEditor"
    DETAILS = "dash:DetailsEditor"
    ENUM_SELECT = "dash:EnumSelectEditor"
    INSTANCES_SELECT = "dash:InstancesSelectEditor"
    RICH_TEXT_EDITOR = "dash:RichTextEditor"
    SUB_CLASS = "dash:SubClassEditor"
    TEXT_AREA = "dash:TextAreaEditor"
    TEXT_AREA_WITH_LANG = "dash:TextAreaWithLangEditor"
    TEXT_FIELD = "dash:TextFieldEditor"
    TEXT_FIELD_WITH_LANG = "dash:TextFieldWithLangEditor"
    URI = "dash:UriEditor"



    @classmethod
    def _missing_(cls, value):
        if not isinstance(value, str):
            return None

        v = value.strip()

        # 1) normalize dash: prefix
        if not v.startswith("dash:"):
            v = f"dash:{v}"

        # 2) try value match (dash:TextAreaEditor)
        for member in cls:
            if member.value == v:
                return member

        # 3) try enum-name match (AUTO_COMPLETE)
        key = v.removeprefix("dash:").replace("-", "_").upper()
        if key in cls.__members__:
            return cls[key]

        return None

    @classmethod
    def get_editor_by_name(cls, name):
        if not name.startswith("dash:"):
            name = f"dash:{name}"
        for editor in cls:
            if editor.name == name:
                return editor
        return None

    @property
    def toRdf(self):
        return self.value

    def _as_dict(self):
        return {"value": self.value }




import json
import os
from copy import deepcopy
from threading import Lock
from typing import Any

import redis

from oldaplib.src.helpers.serializer import serializer
from oldaplib.src.helpers.singletonmeta import SingletonMeta
from oldaplib.src.iconnection import IConnection
from oldaplib.src.xsd.iri import Iri
from oldaplib.src.xsd.xsd_ncname import Xsd_NCName
from oldaplib.src.xsd.xsd_qname import Xsd_QName


class CacheSingleton(metaclass=SingletonMeta):
    """
    Singleton class for thread-safe caching.

    This class provides a mechanism for thread-safe access and modification
    of a cache. It ensures single-instance usage via the SingletonMeta
    metaclass. The cache allows storing, retrieving, deleting, and clearing
    key-value pairs in a thread-safe manner.

    :ivar _lock: Lock object ensuring thread-safe access to the cache.
    :type _lock: Lock
    :ivar _cache: Internal dictionary used for storing the cache data.
    :type _cache: dict[Iri | Xsd_NCName, Any]
    """
    _lock: Lock
    _cache: dict[Iri | Xsd_NCName, Any]

    def __init__(self):
        self._lock = Lock()
        self._cache = {}

    def __str__(self) -> str:
        with self._lock:
            return str(self._cache)

    def get(self, key: Iri | Xsd_NCName) -> Any:
        with self._lock:
            return deepcopy(self._cache.get(key))

    def set(self, key: Iri | Xsd_NCName, value: Any, key2: Iri | Xsd_NCName | None = None) -> None:
        with self._lock:
            self._cache[key] = deepcopy(value)
            if key2 is not None:
                self._cache[key2] = self._cache[key]

    def delete(self, key: Iri | Xsd_NCName):
        with self._lock:
            if key in self._cache:
                self._cache.pop(key, None)

    def clear(self):
        with self._lock:
            self._cache.clear()


class CacheSingletonRedis:
    """
    Singleton class for caching using a Redis database.

    This class interacts with a Redis instance to store, retrieve, and manage
    cached data. Designed to facilitate data caching using key-value pairs,
    supporting serialization and deserialization for complex objects. Provides
    methods for synchronous operations like setting, retrieving, deleting, and
    clearing cache entries.

    :ivar _r: Connection to the Redis database.
    :type _r: redis.client.Redis
    """
    def __init__(self):
        # default connection to local redis server on port 6379

        #self._r = redis.Redis(host=os.getenv("OLDAP_REDIS_HOST", 'localhost'), port=os.getenv("OLDAP_REDIS_PORT", 6379), db=0)

        redis_url = os.getenv("OLDAP_REDIS_URL", "redis://localhost:6379")
        self._r = redis.from_url(redis_url)

    def get(self, key: Iri | Xsd_NCName | Xsd_QName, connection: IConnection | None = None) -> Any:
        value = self._r.get(str(key))
        if connection:
            return json.loads(value, object_hook=serializer.make_decoder_hook(connection=connection)) if value else None
        else:
            return json.loads(value, object_hook=serializer.decoder_hook) if value else None

    def set(self, key: Iri | Xsd_NCName | Xsd_QName, value: Any, key2: Iri | Xsd_NCName | None = None) -> None:
        self._r.set(str(key), json.dumps(value, default=serializer.encoder_default))
        if key2 is not None:
            self._r.set(str(key2), json.dumps(value, default=serializer.encoder_default))

    def delete(self, key: Iri | Xsd_NCName | Xsd_QName):
        self._r.delete(str(key))

    def clear(self):
        self._r.flushdb()

    def exists(self, key: Iri | Xsd_NCName | Xsd_QName) -> bool:
        value = self._r.get(str(key))
        return value is not None



import json
from enum import Enum
from pathlib import Path
from pprint import pprint
from typing import Any

import yamale
import yaml

from oldaplib.src.cachesingleton import CacheSingleton
from oldaplib.src.connection import Connection
from oldaplib.src.helpers.context import Context
from oldaplib.src.helpers.json_encoder import SpecialEncoder
from oldaplib.src.helpers.langstring import LangString
from oldaplib.src.helpers.oldaperror import OldapErrorNotImplemented, OldapErrorValue
from oldaplib.src.helpers.query_processor import QueryProcessor
from oldaplib.src.iconnection import IConnection
from oldaplib.src.oldaplist import OldapList
from oldaplib.src.oldaplistnode import OldapListNode
from oldaplib.src.project import Project
from oldaplib.src.xsd.iri import Iri
from oldaplib.src.xsd.xsd_integer import Xsd_integer
from oldaplib.src.xsd.xsd_ncname import Xsd_NCName

class ListFormat(Enum):
    PYTHON = 'python'
    JSON = 'json'
    YAML = 'yaml'


def get_node_indices(con: IConnection, oldapList: OldapList) -> list[tuple[Iri, Xsd_integer, Xsd_integer]]:
    """
    Retrieve node indices from a given list structure in the ontology.

    This function constructs and executes a SPARQL query to retrieve
    the left and right indices of nodes belonging to the given list
    structure. The result is returned as a list of tuples containing
    the node IRI and its associated indices.

    :param con: An active connection to the ontology system.
    :param oldapList: Representation of a list within the ontology that
        contains contextual and project-specific information.
    :return: A list of tuples containing the node IRI, left index, and
        right index extracted from the ontology list.
    :rtype: list[tuple[Iri, Xsd_integer, Xsd_integer]]
    :raises OldapError: If the connection object is not an instance of IConnection.
    :raises TypeError: If the connection object is not an instance of IConnection.
    :raises ValueError: If the connection object is not an instance of IConnection.
    """
    context = Context(name=con.context_name)
    graph = oldapList.project.projectShortName

    query = context.sparql_context
    query += f"""
    SELECT ?node ?lindex ?rindex
    WHERE {{
        GRAPH {graph}:lists {{
            ?node skos:inScheme {oldapList.iri.toRdf} ;
                oldap:leftIndex ?lindex ;
                oldap:rightIndex ?rindex .
        }}
    }}
    ORDER BY ?node
    """
    jsonobj = con.query(query)
    res = QueryProcessor(context, jsonobj)
    result: list[tuple[Iri, Xsd_integer, Xsd_integer]] = []
    for r in res:
        result.append((r['node'], r['lindex'], r['rindex']))
    return result



def dump_list_to(con: IConnection,
                 project: Project | Iri | Xsd_NCName | str,
                 oldapListId: Xsd_NCName | str,
                 listformat: ListFormat=ListFormat.JSON,
                 ignore_cache=False) -> OldapList | str:
    """
    Dumps an OLDAP list into a specified format such as Python dictionary, JSON,
    or YAML. The function manages and consults a cache to improve performance,
    retrieving the necessary OLDAP list data from either the cache or directly
    from the database, based on user preferences.

    :param con: Connection object `con` used to interact with the database.
                Must provide necessary methods and properties for executing
                backend operations.
    :type con: IConnection

    :param project: Identifies the project to which the OLDAP list belongs.
                    Accepts project-related identifiers such as `Project`,
                    `Iri`, `Xsd_NCName`, or `str`.
    :type project: Project | Iri | Xsd_NCName | str

    :param oldapListId: The identifier for the OLDAP list to dump.
    :type oldapListId: Xsd_NCName | str

    :param listformat: Specifies the output format for the OLDAP list.
                       Supported formats include `ListFormat.JSON`,
                       `ListFormat.PYTHON`, and `ListFormat.YAML`.
                       Defaults to `ListFormat.JSON`.
    :type listformat: ListFormat

    :param ignore_cache: Whether to force fetching the list directly
                         from the database, bypassing any cached
                         value. Defaults to `False`.
    :type ignore_cache: bool

    :return: Depending on the specified `listformat`, returns the OLDAP list
             in one of the supported formats:
             - For `ListFormat.PYTHON`: Returns an `OldapList` instance.
             - For `ListFormat.JSON`: Returns a string representing the list
               as JSON.
             - For `ListFormat.YAML`: Returns a string representing the list
               as YAML.
    :rtype: OldapList | str

    :raises OldapError: If the connection object is not an instance of IConnection.
    :raises TypeError: If the connection object is not an instance of IConnection.
    :raises ValueError: If the connection object is not an instance of IConnection.
    :raises Exception: If an unexpected error occurs.
    """

    def set_con(nodes: list[OldapListNode]) -> None:
        for node in nodes:
            node._con = con
            if node.nodes:
                set_con(node.nodes)

    def make_dict(listnode: OldapList | OldapListNode, listdict: dict) -> None:
        if isinstance(listnode, OldapList):
            listdict[str(listnode.oldapListId)] =  {}
            if listnode.prefLabel:
                listdict[str(listnode.oldapListId)]['label'] = [str(x) for x in listnode.prefLabel]
            if listnode.definition:
                listdict[str(listnode.oldapListId)]['definition'] = [str(x) for x in listnode.definition]
            if listnode.nodes:
                listdict[str(listnode.oldapListId)]['nodes'] = {}
                for node in listnode.nodes:
                    make_dict(node, listdict[str(listnode.oldapListId)]['nodes'])
        else:
            listdict[str(listnode.oldapListNodeId)] =  {}
            if listnode.prefLabel:
                listdict[str(listnode.oldapListNodeId)]['label'] = [str(x) for x in listnode.prefLabel]
            if listnode.definition:
                listdict[str(listnode.oldapListNodeId)]['definition'] = [str(x) for x in listnode.definition]
            if listnode.nodes:
                listdict[str(listnode.oldapListNodeId)]['nodes'] = {}
                for node in listnode.nodes:
                    make_dict(node, listdict[str(listnode.oldapListNodeId)]['nodes'])

    #
    # We need to get the OldapList IRI for asking the cache...
    #
    if not isinstance(project, Project):
        project = Project.read(con, project)
    oldapListIri = Iri.fromPrefixFragment(project.projectShortName, Xsd_NCName(oldapListId), validate=False)
    cache = CacheSingleton()
    listnode = None
    if not ignore_cache:
        listnode = cache.get(oldapListIri)
        if listnode is not None:
            # rectify the connection
            listnode._con = con
            # if listnode.nodes:
            #     set_con(listnode.nodes)
            setattr(listnode, 'source', 'cache')
    if listnode is None:
        #
        # List was not in cache, read it from database
        #
        listnode = OldapList.read(con=con,
                                  project=project,
                                  oldapListId=oldapListId)
        nodes = listnode.nodes
        listnode.nodes = nodes
        setattr(listnode, 'source', 'db')
        cache.set(oldapListIri, listnode)

    match listformat:
        case ListFormat.PYTHON:
            return listnode
        case ListFormat.JSON:
            return json.dumps(listnode, cls=SpecialEncoder, indent=3)
        case ListFormat.YAML:
            list_dict = {}
            make_dict(listnode, list_dict)
            return yaml.dump(list_dict, indent=2, allow_unicode=True)

    return ''

def print_sublist(nodes: list[OldapListNode], level: int = 1) -> None:
    """
    Prints a nested list of OldapListNode objects in a structured format.

    This function recursively traverses through a list of OldapListNode objects,
    and its nested `nodes` attributes, to print their details. Node information
    is displayed with indentation proportional to their depth level in the
    hierarchy.

    :param nodes: A list of OldapListNode objects to be printed. Each node contains
        attributes such as `oldapListNodeId`, `leftIndex`, `rightIndex`,
        `prefLabel`, `iri`, and optionally nested `nodes`.
    :type nodes: list[OldapListNode]
    :param level: The current depth level of indentation. Defaults to 1.
    :type level: int
    :return: None
    """
    for node in nodes:
        print(f'{str(node.oldapListNodeId): >{level * 5}} ({node.leftIndex}, {node.rightIndex}) prefLabel={node.prefLabel} iri={node.iri}')
        if node.nodes:
            print_sublist(node.nodes, level + 1)


def load_list_from_yaml(con: Connection,
                        project: Project | Iri | Xsd_NCName | str,
                        filepath: Path) -> list[OldapList]:
    """
    Loads a list of OldapList objects from a YAML file, validates its structure
    using a predefined schema, and processes its content into objects in the
    system. Each list in the YAML file is mapped to an OldapList instance
    and its nodes recursively.

    :param con: A connection object to interact with the backend system.
    :param project: The project to which the created OldapLists will belong.
        It can be a Project object or identifications such as Iri, Xsd_NCName,
        or string.
    :param filepath: The path of the YAML file containing the lists and nodes
        to be loaded and processed.
    :return: A list of OldapList objects that were successfully created and
        populated based on the YAML input.

    :raises OldapError: If the connection object is not an instance of IConnection.
    :raises TypeError: If the connection object is not an instance of IConnection.
    :raises ValueError: If the connection object is not an instance of IConnection.
    :raises OldapErrorNotImplemented: If the YAML file contains nodes that are not implemented.
    """

    def process_nodes(nodes: dict[str, Any], oldaplist: OldapList, parent: OldapListNode | None):
        oldapnodes: list[OldapListNode] = []
        for nodeid, nodedata in nodes.items():
            if not hasattr(listdata, 'get'):
                raise OldapErrorValue(f'YAML has invalid content.')
            label = LangString(nodedata.get('label'))
            definition = LangString(nodedata.get('definition'))
            node = OldapListNode(con=con,
                                 **oldaplist.info,
                                 oldapListNodeId=Xsd_NCName(nodeid),
                                 prefLabel=label or None,
                                 definition=definition or None)
            if oldapnodes:
                node.insert_node_right_of(oldapnodes[-1])
            else:
                if parent is None:
                    node.create_root_node()
                else:
                    node.insert_node_below_of(parent)
            oldapnodes.append(node)
            if nodedata.get('nodes'):
                node.nodes = process_nodes(nodedata['nodes'], oldaplist, node)
        return oldapnodes

    if not isinstance(project, Project):
        project = Project.read(con, project)

    oldaplists: list[OldapList] = []
    #
    # first we validate the YAML file using the following schema (using yamale)
    #
    with filepath.open() as f:
        schema = yamale.make_schema(content='''
map(include('node'))
---
node:
  label: list(str(matches='^.*@[ -~]{2}$'))
  definition: list(str(matches='^.*@[ -~]{2}$'), required=False)
  nodes: map(include('node'), required=False)
'''
                           )
        data = yamale.make_data(content=f.read())
        try:
            yamale.validate(schema=schema, data=data)
        except ValueError as e:
            raise OldapErrorValue(f"Error validating YAML file: {e}")
    #
    # now we read the YAML
    #
    with filepath.open() as f:
        try:
            lists = yaml.safe_load(f)
        except yaml.YAMLError as e:
            raise OldapErrorValue(f"Error loading YAML file: {e}")
        for listid, listdata in lists.items():
            if not hasattr(listdata, 'get'):
                raise OldapErrorValue(f'YAML has invalid content.')
            label = LangString(listdata.get('label'))
            definition = LangString(listdata.get('definition'))
            oldaplist = OldapList(con=con,
                                  project=project,
                                  oldapListId=Xsd_NCName(listid, validate=True),
                                  prefLabel=label or None,
                                  definition=definition or None,
                                  validate=True)
            oldaplist.create()
            oldaplists.append(oldaplist)
            if listdata.get('nodes'):
                oldaplist.nodes = process_nodes(listdata['nodes'], oldaplist, None)
    return oldaplists

def get_node_by_id(nodes: list[OldapListNode], id: Xsd_NCName) -> OldapListNode | None:
    """
    Retrieve a node from a nested list of OldapListNode objects that matches a given ID.

    This function searches recursively through the list of OldapListNode objects and their
    nested children to find a node with a matching oldapListNodeId. If a match is found,
    the node is returned. If no matching node is found in the entire structure, the function
    returns None.

    :param nodes: A list of OldapListNode objects within which the search is conducted.
    :type nodes: list[OldapListNode]
    :param id: The identifier to match against the oldapListNodeId of nodes.
    :type id: Xsd_NCName
    :return: Returns the matching OldapListNode object if found, or None if no match is located.
    :rtype: OldapListNode | None
    """
    for node in nodes:
        if node.oldapListNodeId == id:
            return node
        elif node.nodes:
            result = get_node_by_id(node.nodes, id)
            if result is not None:
                return result
    return None



"""
# User class

This module implements the Python representation of an OLDAP user. It offers methods to

- create
- read
- search
- update
- delete

a user from the triple store. The User class inherits from
[UserDataclass](/python_docstrings/userdataclass/#oldap.src.user_dataclass.UserFields). UserDataclass
provides the inner workings for the User, but without database access. This separation is necessary
for bootstrapping the connection to the triple store. The [Connection](/python_docstrings/connection) class obviously
has not yet an established connection instance. Therefore, it as to connect to the triple store using basic method,
but still must get authorization based on the supplied userId and credentials. Internally it stores the user's
information including all permissions within the access token that is returned by the first connection (aka login).

The UserDataclass class is serializable as JSON as follows (using the @serializer decorator):
```python
jsonstr = json.dumps(userdata, default=serializer.encoder_default)
user = json.loads(jsonstr, object_hook=serializer.decoder_hook)
```

The User class inherits the following properties from the UserDataclass class:

- _userIri_: IRI of the user, cannot be changed (RDF property `oldap:userIri`)
- _userId_: User ID as NCName (RDF property `oldap:userId`)
- _familyName_: Family name as str (RDF property `foaf:familyName`)
- _givenName_: Given name or first name as str(RDF property `foaf:givenName`)
- _credentials_: Credential (password) (RDF property `oldap:credentials`)
- _isActive_: Is the user active as bool? (RDF property `oldap:isActive`)
- _inProject_: Membership to projects and administrative permissions for this project (RDF property `oldap:inProject)
- _hsPermission_: Permissions for data as sets of QNames (RDF property `oldap:hasPermissions`)

These properties can be accessed as normal python class properties or using the dictionary syntax. The keys
are defined in the [UserFields](/python_docstrings/userdataclass/#oldaplib.src.user_dataclass.UserFields) Enum class.
Example for access as property:
```python
user.familyName = 'Rosenthaler'
givenname = user.givenName
del user.givenName
```
Example for access as dictionary:
```python
user[UserFields.FAMILY_NAME] = 'Rosenthaler'
givenname = user[UserFields.GIVEN_NAME]
del user[UserFields.GIVEN_NAME]
```

## Create a user

A user is created using the method [create()](/python_docstrings/user/#oldaplib.src.user.User.create) as follows:

```python
user = User(con=self._connection,
            userIri=AnyIRI("https://orcid.org/0000-0002-9991-2055"),
            userId=NCName("edison"),
            family_name="Edison",
            given_name="Thomas A.",
            credentials="Lightbulb&Phonograph",
            inProject={QName('oldap:HyperHamlet'): {AdminPermission.ADMIN_USERS,
                                                   AdminPermission.ADMIN_RESOURCES,
                                                   AdminPermission.ADMIN_CREATE}},
            hasPermissions={QName('oldap:GenericView')})
user.create()
```

- __userIri__ is an optional parameter that allows to give the user a unique IRI. If possible, the
  [ORCID](https://orcid.org) ID should be given. If this parameter is omitted, OLDAP generates a unique IRI from the URN
  namespace for the user.
- __userId__ must be an [NCName](/python_docstrings/datatypes#oldaplib.src.helpers.datatypes.NCName)
- __credentials__ is a password that is converted to a bcrypt hash
- __inProject__ is a dictionary with the keys being the projects that the user is a member of. The values are
  sets of administrative privileges as defined in [AdminPermissions](/python_docstrings/permissions#AdminPermissions)
- __hasPermissions__ are links to PermissionSet's that define the access permissions for resources.

Please note that the class constructor does *not* create the user in the triple store. In order to create
the user in the database, `<User>.create()`has to be called.

## Reading a user from the database

In order to read all user data from the triple store, the method [read()](/python_docstrings/user/#oldaplib.src.user.User.read) is used as
follows:

```python
user = User.read(con=self._connection, userId="rosenth")
```

The `userId` must be known and passed either as string or NCName.

## Searching for a user in the database

OLDAP allows to search for users within the database. The method [search()](/python_docstrings/user/#oldaplib.src.user.User.search)
performs a search. The string given must match in total with the entry in the database. The method accepts also
several arguments which are combined by a logical AND.

```python
users = User.search(con=self._connection,userId="fornaro")

users = User.search(con=self._connection, familyName="Rosenthaler")

users = User.search(con=self._connection, givenName="John")

users = User.search(con=self._connection, inProject=QName("oldap:HyperHamlet"))

users = User.search(con=self._connection, inProject=AnyIRI("http://www.salsah.org/version/2.0/SwissBritNet"))

users = User.search(con=self._connection, userId="GAGA")
```

## Updating a User

Several properties of a user can be changed using the [update()](/python_docstrings/user/#oldaplib.src.user.User.update) method. In a first step, the properties
of a user instance are changed, then the `update()` method writes the changes to the triple store.

The following example exemplifies the procedure:

```python
    user2 = User.read(con=self._connection, userId="edison")  # read the user from the triple store
    user2.userId = "aedison"  # change the userId
    user2.familyName = "Edison et al."  # change the familyName
    user2.givenName = "Thomas"  # change the givenName
    user2.hasPermissions.add(QName('oldap:GenericRestricted'))  # add a permission set
    user2.hasPermissions.add(QName('oldap:HyperHamletMember'))  # add a permission set
    user2.hasPermissions.remove(QName('oldap:GenericView'))  # remove a permission set
    user2.inProject[QName('oldap:SystemProject')] = {AdminPermission.ADMIN_USERS, AdminPermission.ADMIN_RESOURCES}
    user2.inProject[QName('oldap:HyperHamlet')].remove(AdminPermission.ADMIN_USERS)
    user2.update()
```

## Deleting a User

The method [delete()](/python_docstrings/user/#oldaplib.src.user.User.delete) deletes the given user from
the database:

```python
user3 = User.read(con=self._connection, userId="aedison")
user3.delete()
```
"""
import textwrap
from copy import deepcopy
from enum import Enum
from functools import partial
from pprint import pprint
from typing import List, Self, Optional, Any, Tuple

import bcrypt

from oldaplib.src.cachesingleton import CacheSingleton, CacheSingletonRedis
from oldaplib.src.enums.action import Action
from oldaplib.src.enums.datapermissions import DataPermission
from oldaplib.src.enums.userattr import UserAttr
from oldaplib.src.helpers.context import Context
from oldaplib.src.helpers.irincname import IriOrNCName
from oldaplib.src.helpers.observable_dict import ObservableDict
from oldaplib.src.helpers.observable_set import ObservableSet
from oldaplib.src.helpers.serializer import serializer
from oldaplib.src.in_project import InProjectClass
from oldaplib.src.userdataclass import UserData
from oldaplib.src.xsd.iri import Iri
from oldaplib.src.xsd.xsd_anyuri import Xsd_anyURI
from oldaplib.src.xsd.xsd_qname import Xsd_QName
from oldaplib.src.xsd.xsd_ncname import Xsd_NCName
from oldaplib.src.xsd.xsd_datetime import Xsd_dateTime
from oldaplib.src.xsd.xsd_string import Xsd_string
from oldaplib.src.helpers.oldaperror import OldapError, OldapErrorAlreadyExists, OldapErrorNotFound, OldapErrorUpdateFailed, \
    OldapErrorValue, OldapErrorNoPermission
from oldaplib.src.helpers.query_processor import QueryProcessor
from oldaplib.src.enums.adminpermissions import AdminPermission
from oldaplib.src.iconnection import IConnection
from oldaplib.src.model import Model
from oldaplib.src.helpers.attributechange import AttributeChange


@serializer
class User(Model):
    """
    This class represents a user in the OLDAP system. It allows managing and handling user-related
    functionalities, including working with permissions, projects, and credentials. The class relies
    on the provided connection and other related attributes for interaction with the triple store.
    It also incorporates the `InProjectClass` and `ObservableSet` to handle project and permission
    modifications.

    :ivar userIri: The unique identifier (IRI) for the user. Automatically created if not provided.
    :type userIri: Iri
    :ivar inProject: A mapping of project identifiers to their respective sets of permissions.
        Modifications trigger callbacks.
    :type inProject: InProjectClass
    :ivar hasPermissions: Permissions assigned to the user. Supports modification callbacks.
    :type hasPermissions: ObservableSet
    :ivar credentials: The user’s hashed credentials for secure authentication.
    :type credentials: Xsd_string
    """

    def __init__(self, *,
                 con: IConnection,
                 creator: Iri | str | None = None,
                 created: Xsd_dateTime | str | str | None = None,
                 contributor: Iri | str | None = None,
                 modified: Xsd_dateTime | str | None = None,
                 validate: bool = False,
                 **kwargs):
        """
        Initializes a user instance with specified attributes and performs consistency checks for the user.

        :param con: The connection instance used for database or API interactions.
        :type con: IConnection

        :param creator: The entity or identifier for the creator of the user, provided as an Iri or string.
        :type creator: Iri | str | None

        :param created: The datetime information of the creation moment, provided as an Xsd_dateTime,
                        string, or None.
        :type created: Xsd_dateTime | str | None

        :param contributor: The entity or identifier for the contributor, provided as an Iri or string.
        :type contributor: Iri | str | None

        :param modified: The datetime information for when the user was last modified, provided as an
                         Xsd_dateTime, string, or None.
        :type modified: Xsd_dateTime | str | None

        :param validate: A boolean flag indicating whether to perform validation checks on the user instance.
        :type validate: bool

        :param kwargs: Additional keyword arguments that specify custom attributes for the user.

        :raises OldapErrorValue: If the user ID is invalid.
        """
        super().__init__(connection=con,
                         created=created,
                         creator=creator,
                         modified=modified,
                         contributor=contributor,
                         validate=validate)

        self.set_attributes(kwargs, UserAttr)

        #
        # maybe we have to change the default data permissionss for the user
        #
        if self._attributes.get(UserAttr.HAS_ROLE):
            tmp = ObservableDict()
            for role, dperm in self._attributes[UserAttr.HAS_ROLE].items():
                if not isinstance(dperm, DataPermission):
                    tmp[role] = DataPermission.from_qname(dperm) if isinstance(dperm, Xsd_QName) else DataPermission.from_string(dperm)
                else:
                    tmp[role] = dperm
            self._attributes[UserAttr.HAS_ROLE] = tmp
        #
        # Consistency checks
        #
        if not self._attributes.get(UserAttr.USER_IRI):
            self._attributes[UserAttr.USER_IRI] = Iri()  # create URN as userIri

        if self._attributes.get(UserAttr.IN_PROJECT):
            self._attributes[UserAttr.IN_PROJECT].set_notifier(self.__inProject_cb, UserAttr.IN_PROJECT)
        else:
            self._attributes[UserAttr.IN_PROJECT] = InProjectClass(notifier=self.__inProject_cb)

        if self._attributes.get(UserAttr.HAS_ROLE):
            #self._attributes[UserAttr.HAS_ROLE].set_on_change(self.__hasRole_cb, UserAttr.HAS_ROLE)
            self._attributes[UserAttr.HAS_ROLE].set_on_change(self.__hasRole_cb)
        else:
            self._attributes[UserAttr.HAS_ROLE] = ObservableDict(on_change=self.__hasRole_cb)

        if self._attributes.get(UserAttr.CREDENTIALS):
            if not str(self._attributes[UserAttr.CREDENTIALS]).startswith(('$2a$', '$2b$', '$2y$')):
                salt = bcrypt.gensalt()
                self._attributes[UserAttr.CREDENTIALS] = Xsd_string(bcrypt.hashpw(str(self._attributes[UserAttr.CREDENTIALS]).encode('utf-8'), salt).decode('utf-8'))

        for attr in UserAttr:
            setattr(User, attr.value.fragment, property(
                partial(User._get_value, attr=attr),
                partial(User._set_value, attr=attr),
                partial(User._del_value, attr=attr)))
        self.clear_changeset()

    def _as_dict(self):
        return {x.fragment: y for x, y in self._attributes.items()} | super()._as_dict()

    def __deepcopy__(self, memo: dict[Any, Any]) -> Self:
        if id(self) in memo:
            return memo[id(self)]
        cls = self.__class__
        instance = cls.__new__(cls)
        memo[id(self)] = instance
        Model.__init__(instance,
                       connection=deepcopy(self._con, memo),
                       creator=deepcopy(self._creator, memo),
                       created=deepcopy(self._created, memo),
                       contributor=deepcopy(self._contributor, memo),
                       modified=deepcopy(self._modified, memo))
        # Copy internals of Model:
        instance._attributes = deepcopy(self._attributes, memo)
        instance._changset = deepcopy(self._changeset, memo)
        return instance

    def cleanup_setter(self, attr: UserAttr, value: Any):
        if attr == UserAttr.CREDENTIALS:
            salt = bcrypt.gensalt()
            self._attributes[UserAttr.CREDENTIALS] = Xsd_string(bcrypt.hashpw(str(value).encode('utf-8'), salt).decode('utf-8'))
        if attr == UserAttr.IN_PROJECT:
            if value is None:
                self._attributes[UserAttr.IN_PROJECT] = InProjectClass(notifier=self.__inProject_cb)
            else:
                self._attributes[UserAttr.IN_PROJECT] = InProjectClass(value, notifier=self.__inProject_cb)
        if attr == UserAttr.HAS_ROLE:
            if value is None:
                self._attributes[UserAttr.HAS_ROLE] = ObservableDict(on_change=self.__hasRole_cb)
            else:
                self._attributes[UserAttr.HAS_ROLE] = ObservableDict(value, on_change=self.__hasRole_cb)

    def check_for_permissions(self) -> Tuple[bool, str]:
        """
        Evaluates whether the currently logged-in user ("actor") has the necessary permissions
        to create a user for the specified project(s). The function checks if the actor has root
        privileges or ADMIN_USERS permission for each relevant project. Returns appropriate
        messages indicating the state of permissions.

        :returns: A tuple where the first element is a boolean indicating whether the actor has
                  permissions, and the second element is a string message providing details.
        :rtype: tuple[bool, str]
        """
        #
        # First we check if the logged-in user ("actor") has the permission to create a user for
        # the given project!
        #
        actor = self._con.userdata
        sysperms = actor.inProject.get(Iri('oldap:SystemProject'))
        if sysperms and AdminPermission.ADMIN_OLDAP in sysperms:
            #
            # user has root privileges!
            #
            return True, "OK – IS ROOT"
        else:
            if not self.inProject:
                return False, f'Actor has no ADMIN_USERS permission for user {self.userId}.'
            allowed: list[Iri] = []
            for proj in self.inProject.keys():
                if actor.inProject.get(proj) is None:
                    return False, f'Actor has no ADMIN_USERS permission for project {proj}'
                else:
                    if AdminPermission.ADMIN_USERS not in actor.inProject.get(proj):
                        return False, f'Actor has no ADMIN_USERS permission for project {proj}'
            return True, "OK..."

    #
    # Callbacks for the `ObservableSet`class. This is used whenever the `hasPermission`or
    # `inProject`properties are being modified
    #
    def __hasRole_cb(self, old_value: ObservableDict) -> None:
        if self._changeset.get(UserAttr.HAS_ROLE) is None:
            self._changeset[UserAttr.HAS_ROLE] = AttributeChange(old_value, Action.MODIFY)

    def __inProject_cb(self, data: Enum | Iri = None) -> None:
        if self._changeset.get(UserAttr.IN_PROJECT) is None:
            self._changeset[UserAttr.IN_PROJECT] = AttributeChange(None, Action.MODIFY)

    def add_project_permission(self, project: Iri | str, permission: AdminPermission | None) -> None:
        if self._attributes[UserAttr.IN_PROJECT].get(project) is None:
            if self._changeset.get(UserAttr.IN_PROJECT) is None:
                self._changeset[UserAttr.IN_PROJECT] = AttributeChange(self._attributes[UserAttr.IN_PROJECT], Action.CREATE)
            self._attributes[UserAttr.IN_PROJECT][project] = ObservableSet({permission})
        else:
            if self._changeset.get(UserAttr.IN_PROJECT) is None:
                self._changeset[UserAttr.IN_PROJECT] = AttributeChange(self._attributes[UserAttr.IN_PROJECT], Action.MODIFY)
            self._attributes[UserAttr.IN_PROJECT][project].add(permission)

    def remove_project_permission(self, project: Iri | str, permission: AdminPermission | None) -> None:
        if not isinstance(project, Iri):
            project = Iri(project)
        if self._attributes[UserAttr.IN_PROJECT].get(project) is None:
            raise OldapErrorValue(f"Project '{project}' does not exist")
        if self._changeset.get(UserAttr.IN_PROJECT) is None:
            self._changeset[UserAttr.IN_PROJECT] = AttributeChange(self._attributes[UserAttr.IN_PROJECT], Action.MODIFY)
        self._attributes[UserAttr.IN_PROJECT][project].remove(permission)

    def trig_to_str(self,
                    creator: Iri,
                    created: Xsd_dateTime,
                    contributor: Iri,
                    modified: Xsd_dateTime,
                    indent: int = 0, indent_inc: int = 4):
        blank = ''
        sparql = ''
        sparql += f'{blank:{indent * indent_inc}}{self.userIri.toRdf} a oldap:User'
        sparql += f' ;\n{blank:{(indent + 1) * indent_inc}}dcterms:creator {creator.toRdf}'
        sparql += f' ;\n{blank:{(indent + 1) * indent_inc}}dcterms:created {created.toRdf}'
        sparql += f' ;\n{blank:{(indent + 1) * indent_inc}}dcterms:contributor {contributor.toRdf}'
        sparql += f' ;\n{blank:{(indent + 1) * indent_inc}}dcterms:modified {modified.toRdf}'
        sparql += f' ;\n{blank:{(indent + 1) * indent_inc}}oldap:userId {self.userId.toRdf}'
        sparql += f' ;\n{blank:{(indent + 1) * indent_inc}}schema:familyName {self.familyName.toRdf}'
        sparql += f' ;\n{blank:{(indent + 1) * indent_inc}}schema:givenName {self.givenName.toRdf}'
        sparql += f' ;\n{blank:{(indent + 1) * indent_inc}}schema:email {self.email.toRdf}'
        sparql += f' ;\n{blank:{(indent + 1) * indent_inc}}oldap:credentials {self.credentials.toRdf}'
        activeval = "true" if self.isActive else "false"
        sparql += f' ;\n{blank:{(indent + 1) * indent_inc}}oldap:isActive {activeval}'
        star = ''
        if self.inProject:
            project = [p.toRdf for p in self.inProject.keys()]
            rdfstr = ", ".join(project)
            sparql += f' ;\n{blank:{(indent + 1) * indent_inc}}oldap:inProject {rdfstr}'
            for p in self.inProject.keys():
                for admin_p in self.inProject[p]:  # TODO: May be use .get() instead of [] !!!!!!!!!!!!!!!!!!!!!!!!!
                    star += f'{blank:{indent * indent_inc}}<<{self.userIri.toRdf} oldap:inProject {p.toRdf}>> oldap:hasAdminPermission {admin_p.value} .\n'
        if self.hasRole:
            rdfstr = ", ".join([str(x) for x in self.hasRole.keys()])
            sparql += f' ;\n{blank:{(indent + 1) * indent_inc}}oldap:hasRole {rdfstr}'
            for r, p in self.hasRole.items():
                if p:
                    star += f'{blank:{(indent + 1) * indent_inc}}<<{self.userIri.toRdf} oldap:hasRole {r.toRdf}>> oldap:hasDefaultDataPermission {p.value} .\n'
        sparql += " .\n\n"
        sparql += star
        return sparql


    def create(self, indent: int = 0, indent_inc: int = 4, keep_dates: bool = False) -> None:
        """
        Creates a user in the triple store with the provided details. Before proceeding with the
        creation, it verifies the uniqueness of the `userId` and `userIri`. The user is created
        along with its associated attributes and permissions. Necessary checks are performed
        to ensure the provided projects and permission sets exist.

        :param indent: The current indentation level for the generated SPARQL query.
        :param indent_inc: Increment used for indentation in the SPARQL query.
        :return: None
        :raises OldapErrorAlreadyExists: Raised when a user with the same `userId` or `userIri`
            already exists.
        :raises OldapValueError: Raised when a provided project or permission set does not exist.
        :raises OldapError: Raised for internal errors during the user creation process.
        :raises OldapErrorNoPermission: Raised when permissions to create a user for the specified
            project(s) are lacking.
        """
        if self._con is None:
            raise OldapError("Cannot create: no connection")

        result, message = self.check_for_permissions()
        if not result:
            raise OldapErrorNoPermission(message)

        if self.userIri is None:
            self.userIri = Iri()
        context = Context(name=self._con.context_name)
        sparql1 = context.sparql_context
        #
        # SPARQL to test if a user with the same userId already exists
        #
        sparql1 += textwrap.dedent(f"""
        ASK
        FROM oldap:admin
        WHERE {{
            ?user oldap:userId {self.userId.toRdf} .
        }}
        """)

        #
        # SPARQL to test if a user with the same userIri already exists
        #
        sparql2 = context.sparql_context
        sparql2 += textwrap.dedent(f"""
        ASK
        FROM oldap:admin
        WHERE {{
             {self.userIri.toRdf} ?p ?o
        }}
        """)

        #
        # SPARQL to test if the projects the user should be in do exist...
        #
        proj_test = None
        if self.inProject:
            projs = [x.toRdf for x in self.inProject.keys()]
            projslist = ", ".join(projs)
            proj_test = context.sparql_context
            proj_test += textwrap.dedent(f"""
            SELECT ?project
            FROM oldap:admin
            WHERE {{
                ?project a oldap:Project .
                FILTER(?project IN ({projslist}))
            }}
            """)
        else:
            projs = []

        #
        # SPARQL to test if the roles the user should have do exist...
        #
        roles_test = None
        if self.hasRole:
            roles = [x.toRdf for x in self.hasRole]
            roles = ", ".join(roles)
            roles_test = context.sparql_context
            roles_test += textwrap.dedent(f"""
            SELECT ?role
            FROM oldap:admin
            WHERE {{
                ?role a oldap:Role .
                FILTER(?role IN ({roles}))
            }}
            """)
        else:
            roles = []

        timestamp = Xsd_dateTime.now()
        blank = ''
        sparql = context.sparql_context
        sparql += f'{blank:{indent * indent_inc}}INSERT DATA {{\n'
        sparql += f'{blank:{(indent + 1) * indent_inc}}GRAPH oldap:admin {{\n'

        if keep_dates:
            sparql += self.trig_to_str(creator=self.creator,
                                       created=self.created,
                                       contributor=self.contributor,
                                       modified=self.modified,
                                       indent=indent + 2, indent_inc=indent_inc)
        else:
            sparql += self.trig_to_str(creator=self._con.userIri,
                                       created=timestamp,
                                       contributor=self._con.userIri,
                                       modified=timestamp,
                                       indent=indent + 2, indent_inc=indent_inc)
        sparql += f'{blank:{(indent + 1) * indent_inc}}}}\n'
        sparql += f'{blank:{indent * indent_inc}}}}\n'

        self._con.transaction_start()
        try:
            jsonobj = self._con.transaction_query(sparql1)
        except OldapError:
            self._con.transaction_abort()
            raise
        if jsonobj['boolean']:
            self._con.transaction_abort()
            raise OldapErrorAlreadyExists(f'A user with a user ID "{self.userId}" already exists')

        try:
            jsonobj = self._con.transaction_query(sparql2)
        except OldapError:
            self._con.transaction_abort()
            raise
        if jsonobj['boolean']:
            self._con.transaction_abort()
            raise OldapErrorAlreadyExists(f'A user with a user IRI "{self.userIri}" already exists')

        if self.inProject:
            try:
                jsonobj = self._con.transaction_query(proj_test)
            except OldapError:
                self._con.transaction_abort()
                raise
            res = QueryProcessor(context, jsonobj)
            if len(res) != len(projs):
                self._con.transaction_abort()
                raise OldapErrorValue("One of the projects is not existing!")

        if self.hasRole:
            try:
                jsonobj = self._con.transaction_query(roles_test)
            except OldapError:
                self._con.transaction_abort()
                raise
            res = QueryProcessor(context, jsonobj)
            if len(res) != len(self.hasRole):
                self._con.transaction_abort()
                raise OldapErrorValue(f"One of the Roles {self.hasRole} is not existing!")

        try:
            self._con.transaction_update(sparql)
        except OldapError:
            self._con.transaction_abort()
            raise
        try:
            self._con.transaction_commit()
        except OldapError:
            self._con.transaction_abort()
            raise
        self._creator = self._con.userIri
        self._created = timestamp
        self._contributor = self._con.userIri
        self._modified = timestamp
        cache = CacheSingletonRedis()
        cache.set(self.userIri, self)


    @classmethod
    def read(cls,
             con: IConnection,
             userId: IriOrNCName | str,
             ignore_cache: bool = False) -> Self:
        """
        Reads a User instance from the data in the triple store.

        This method queries the triple store for a User instance corresponding
        to the provided `userId`. It utilizes an optional cache to improve
        performance, bypassing the query if the data is already cached. If no
        data is found in either the cache or the triple store, an exception
        is raised.

        :param con: The connection object used to query the triple store.
        :type con: IConnection
        :param userId: The identifier for the user to be retrieved.
        :type userId: IriOrNCName | str
        :param ignore_cache: Controls whether the method should skip checking
                             the cache and query the triple store directly.
                             Defaults to False.
        :type ignore_cache: bool
        :return: An instance of the class populated with user data from the
                 triple store.
        :rtype: Self
        :raises OldapErrorNotFound: Raised when the requested user does not
                                    exist in the triple store.
        """
        if not isinstance(userId, IriOrNCName):
            userId = IriOrNCName(userId, validate=True)
        user_id, user_iri = userId.value()
        if user_iri is not None:
            if not ignore_cache:
                cache = CacheSingletonRedis()
                tmp = cache.get(user_iri, connection=con)
                if tmp is not None:
                    return tmp

        context = Context(name=con.context_name)
        jsonobj = con.query(UserData.sparql_query(context, userId))
        res = QueryProcessor(context, jsonobj)
        if len(res) == 0:
            raise OldapErrorNotFound(f'User "{userId}" not found.')
        userdata = UserData.from_query(res)
        if userdata.inProject:
            userdata.inProject.set_notifier(cls.__inProject_cb, UserAttr.IN_PROJECT)
        instance = cls(con=con,
                       creator=userdata.creator,
                       created=userdata.created,
                       contributor=userdata.contributor,
                       modified=userdata.modified,
                       userIri=userdata.userIri,
                       userId=userdata.userId,
                       familyName=userdata.familyName,
                       givenName=userdata.givenName,
                       email=userdata.email,
                       credentials=userdata.credentials,
                       isActive=userdata.isActive,
                       inProject=userdata.inProject,
                       hasRole=userdata.hasRole)
        cache = CacheSingletonRedis()
        cache.set(instance.userIri, instance)
        instance.clear_changeset()
        return instance

    @staticmethod
    def search(*, con: IConnection,
               userId: Optional[Xsd_NCName | str] = None,
               familyName: Optional[str | Xsd_string] = None,
               givenName: Optional[str | Xsd_string] = None,
               inProject: Optional[Iri | str] = None) -> List[Xsd_anyURI]:
        """
        Search for a user in the database. The search can be refined by specifying one or
        more of the following parameters:

        - `userId`: The unique identifier of the user.
        - `familyName`: The user's last name.
        - `givenName`: The user's first name.
        - `inProject`: The project in which the user is a member.

        Each provided parameter will filter the search results. If multiple parameters are
        specified, the search results will match all specified criteria (logical AND operation).

        :param con: Connection instance for accessing the database.
        :type con: IConnection
        :param userId: The unique identifier of the user to be searched (optional).
        :type userId: Xsd_NCName | str
        :param familyName: The last name of the user to be searched (optional).
        :type familyName: str | Xsd_string
        :param givenName: The first name of the user to be searched (optional).
        :type givenName: str | Xsd_string
        :param inProject: A project identifier to match users who are members of the project (optional).
        :type inProject: Iri | str
        :return: A list of IRIs corresponding to users matching the search criteria.
        :rtype: List[Xsd_anyURI]
        """
        if userId and not isinstance(userId, Xsd_NCName):
            userId = Xsd_NCName(userId, validate=True)
        familyName = Xsd_string(familyName, validate=True) if familyName else None
        givenName = Xsd_string(givenName, validate=True) if givenName else None
        if not isinstance(inProject, Iri):
            inProject = Iri(inProject, validate=True) if inProject else None
        context = Context(name=con.context_name)
        sparql = context.sparql_context
        sparql += 'SELECT DISTINCT ?user\n'
        sparql += 'FROM oldap:onto\n'
        sparql += 'FROM shared:onto\n'
        sparql += 'FROM NAMED oldap:admin\n'
        sparql += 'WHERE {\n'
        sparql += '   GRAPH oldap:admin {\n'
        sparql += '       ?user a oldap:User .\n'
        if userId is not None:
            sparql += '       ?user oldap:userId ?user_id .\n'
            sparql += f'       FILTER(?user_id = {userId.toRdf})\n'
        if familyName is not None:
            sparql += '       ?user schema:familyName ?family_name .\n'
            sparql += f'       FILTER(STR(?family_name) = {familyName.toRdf})\n'
        if givenName is not None:
            sparql += '       ?user schema:givenName ?given_name .\n'
            sparql += f'       FILTER(STR(?given_name) = {givenName.toRdf})\n'
        if inProject is not None:
            sparql += '       ?user oldap:inProject ?project .\n'
            sparql += f'      FILTER(?project = {inProject.toRdf})\n'
        sparql += '    }\n'
        sparql += '}\n'
        jsonobj = con.query(sparql)
        res = QueryProcessor(context, jsonobj)
        users = []
        for r in res:
            users.append(r['user'])
        return users

    def delete(self) -> None:
        """
        Deletes the current user from the triple store. This method ensures that the user is
        removed from the system and any context where the user is associated. Additionally,
        it includes permission checks prior to deletion, and it also considers potential references
        to the user (such as owning data), which may cause the user to be set inactive instead
        of outright deletion.

        :raises OldapErrorNoPermission: If the user does not have the necessary permissions to
                                         perform the delete action.
        :return: None
        """
        result, message = self.check_for_permissions()
        if not result:
            raise OldapErrorNoPermission(message)

        #
        # TODO: Test, if the User is referenced as Owner of data etc. If so, raise an error. The User should then
        # be set inactive by setting the flag "isActive" to False!
        #

        context = Context(name=self._con.context_name)
        sparql = context.sparql_context
        sparql += f"""
        WITH oldap:admin
        DELETE {{
            <<?user oldap:inProject ?proj>> oldap:hasAdminPermission ?rval .    
        }}
        WHERE {{
            ?user a oldap:User .
            ?user oldap:userId {self.userId.toRdf} .
        }} ;
        WITH oldap:admin
        DELETE {{
            <<?user oldap:hasRole ?role>> oldap:hasDefaultDataPermission ?dval .    
        }}
        WHERE {{
            ?user a oldap:User .
            ?user oldap:userId {self.userId.toRdf} .
        }} ;
        DELETE WHERE {{
            ?user a oldap:User .
            ?user oldap:userId {self.userId.toRdf} .
            ?user ?prop ?val .
        }} 
        """
        # TODO: use transaction for error handling
        self._con.update_query(sparql)
        cache = CacheSingletonRedis()
        cache.delete(self.userIri)


    def update(self, indent: int = 0, indent_inc: int = 4) -> None:
        """
        Update user data in the triple store. This method applies all modifications made
        to the user object and updates the database accordingly. It manages changes in
        fields, permissions, and associated projects, ensuring data consistency while
        also checking necessary conditions and permissions.

        This method generates SPARQL queries to handle the updates, and executes them to
        persist changes in the backend. Specific logic handles special cases such as
        updating user permissions, managing projects where the user is involved, and
        dealing with potential race conditions.

        :param indent: The base indent for the generated SPARQL queries.
        :type indent: int
        :param indent_inc: Incremental indent step for improved query readability.
        :type indent_inc: int
        :return: None
        :rtype: None
        :raises OldapErrorUpdateFailed: If the update fails due to a race condition.
        :raises OldapValueError: If a referenced PermissionSet is invalid or does not exist.
        :raises OldapError: If an internal, unspecified error occurs.
        """
        if self._con is None:
            raise OldapError("Cannot create: no connection")

        result, message = self.check_for_permissions()
        if not result:
            #
            # Special case: The user should be able to change its own password!
            #
            actor = self._con.userdata
            if actor.userIri == self.userIri and len(self.changeset) == 1 and self.changeset.get(UserAttr.CREDENTIALS):
                result = True
            if not result:
                raise OldapErrorNoPermission(message)

        timestamp = Xsd_dateTime.now()
        context = Context(name=self._con.context_name)

        ptest = None
        ptest_len = 0
        proj_test = None
        proj_test_len = 0

        blank = ''
        sparql_list = []
        for field, change in self._changeset.items():
            if field == UserAttr.HAS_ROLE or field == UserAttr.IN_PROJECT:
                continue
            sparql = f'{blank:{indent * indent_inc}}# User field "{field.value}" with action "{change.action.value}"\n'
            sparql += f'{blank:{indent * indent_inc}}WITH oldap:admin\n'
            if change.action != Action.CREATE:
                sparql += f'{blank:{indent * indent_inc}}DELETE {{\n'
                sparql += f'{blank:{(indent + 1) * indent_inc}}?user {field.value} {change.old_value.toRdf} .\n'
                sparql += f'{blank:{indent * indent_inc}}}}\n'
            if change.action != Action.DELETE:
                sparql += f'{blank:{indent * indent_inc}}INSERT {{\n'
                sparql += f'{blank:{(indent + 1) * indent_inc}}?user {field.value} {self._attributes[field].toRdf} .\n'
                sparql += f'{blank:{indent * indent_inc}}}}\n'
            sparql += f'{blank:{indent * indent_inc}}WHERE {{\n'
            sparql += f'{blank:{(indent + 1) * indent_inc}}BIND({self.userIri.toRdf} as ?user)\n'
            if change.action != Action.CREATE:
                sparql += f'{blank:{(indent + 1) * indent_inc}}?user {field.value} {change.old_value.toRdf} .\n'
            sparql += f'{blank:{indent * indent_inc}}}}'
            sparql_list.append(sparql)

        if UserAttr.HAS_ROLE in self._changeset:
            added = {}
            removed = {}
            changed = {}
            if self._changeset[UserAttr.HAS_ROLE].action == Action.CREATE:
                added = self._attributes[UserAttr.HAS_ROLE]
            elif self._changeset[UserAttr.HAS_ROLE].action == Action.DELETE:
                removed = self._changeset[UserAttr.HAS_ROLE].old_value
            elif self._changeset[UserAttr.HAS_ROLE].action == Action.REPLACE:
                added = self._attributes[UserAttr.HAS_ROLE]
                removed = self._changeset[UserAttr.HAS_ROLE].old_value
            elif self._changeset[UserAttr.HAS_ROLE].action == Action.MODIFY:
                if self._changeset[UserAttr.HAS_ROLE].old_value:
                    added = {key: self._attributes[UserAttr.HAS_ROLE][key] for key in self._attributes[UserAttr.HAS_ROLE].keys() - self._changeset[UserAttr.HAS_ROLE].old_value.keys()}
                else:
                    added = self._attributes[UserAttr.HAS_ROLE]
                if self._changeset[UserAttr.HAS_ROLE].old_value:
                    removed = {key: self._changeset[UserAttr.HAS_ROLE].old_value[key] for key in self._changeset[UserAttr.HAS_ROLE].old_value.keys() - self._attributes[UserAttr.HAS_ROLE].keys()}
                else:
                    removed = {}
                if self._changeset[UserAttr.HAS_ROLE].old_value:
                    changed = {key: {'old': self._changeset[UserAttr.HAS_ROLE].old_value.get(key), 'new': self._attributes[UserAttr.HAS_ROLE].get(key)}
                               for key in self._changeset[UserAttr.HAS_ROLE].old_value.keys() & self._attributes[UserAttr.HAS_ROLE].keys()
                               if self._attributes[UserAttr.HAS_ROLE][key] != self._changeset[UserAttr.HAS_ROLE].old_value[key]}
                else:
                    changed = {}

            #
            # first we add new roles or delete removed roles from the user (oldap:hasRole). Roles where the
            # oldap:hasDefaultDataPermission has changed are not considered here (but below) since its a RDF*star thing!
            #
            sparql = f'{blank:{indent * indent_inc}}# User field "hasRole"\n'
            sparql += f'{blank:{indent * indent_inc}}WITH oldap:admin\n'
            if removed:
                sparql += f'{blank:{indent * indent_inc}}DELETE {{\n'
                for role in removed.keys():
                    sparql += f'{blank:{(indent + 1) * indent_inc}}?user oldap:hasRole {role.toRdf} .\n'
                sparql += f'{blank:{indent * indent_inc}}}}\n'
            if added:
                sparql += f'{blank:{indent * indent_inc}}INSERT {{\n'
                for role in added.keys():
                    sparql += f'{blank:{(indent + 1) * indent_inc}}?user oldap:hasRole {role.toRdf} .\n'
                sparql += f'{blank:{indent * indent_inc}}}}\n'
            sparql += f'{blank:{indent * indent_inc}}WHERE {{\n'
            sparql += f'{blank:{(indent + 1) * indent_inc}}BIND({self.userIri.toRdf} as ?user)\n'
            sparql += f'{blank:{(indent + 1) * indent_inc}}?user a oldap:User .\n'
            sparql += f'{blank:{indent * indent_inc}}}}'
            if removed or added:
                sparql_list.append(sparql)

            #
            # Now we process the RDF*star triples where a Role has been added/deleted or the
            # oldap:hasDefaultDataPermission has changed!
            #
            rdfstar = False
            sparql = ''
            if removed or changed:
                sparql += f'{blank:{indent * indent_inc}}# RDF*Star DELETE: <<.. oldap:hasRole ...>> oldap:hasDefaultDataPermission ...\n'
                sparql += f'{blank:{indent * indent_inc}}DELETE DATA {{\n'
                sparql += f'{blank:{(indent + 1) * indent_inc}}GRAPH oldap:admin {{\n'
                rdfstar = True
            if removed:
                tmp = [role for role, dperm in removed.items() if dperm] # check if we have roles with dterm not None
                if tmp:
                    for role, dperm in removed.items():
                        if dperm:
                            sparql += f'{blank:{(indent + 2) * indent_inc}}<<{self.userIri.toRdf} oldap:hasRole {role.toRdf}>> oldap:hasDefaultDataPermission {dperm.toRdf} .\n'
            if changed:  # remove RDF*Star triples of the roles that have changed
                tmp = [role for role, dperm in changed.items() if dperm['old']]
                if tmp:
                    for role, changes in changed.items():
                        if changes['old']:
                            sparql += f'{blank:{(indent + 1) * indent_inc}}<<{self.userIri.toRdf} oldap:hasRole {role.toRdf}>> oldap:hasDefaultDataPermission {changes["old"].toRdf} .\n'

            if rdfstar:
                sparql += f'{blank:{(indent + 1) * indent_inc}}}}\n'
                sparql += f'{blank:{indent * indent_inc}}}}\n'
            if added or changed:
                if rdfstar:
                    sparql += f'{blank:{indent * indent_inc}};\n'
                sparql += f'{blank:{indent * indent_inc}}# RDF*Star INSERT: <<.. oldap:hasRole ...>> oldap:hasDefaultDataPermission ...\n'
                sparql += f'{blank:{indent * indent_inc}}INSERT DATA {{\n'
                sparql += f'{blank:{(indent + 1) * indent_inc}}GRAPH oldap:admin {{\n'
                rdfstar = True
            else:
                rdfstar = False
            if added:
                tmp = [role for role, dperm in added.items() if dperm]  # check if we have roles with dterm not None
                if tmp:
                    for role, dperm in added.items():
                        if dperm:
                            sparql += f'{blank:{(indent + 2) * indent_inc}}<<{self.userIri.toRdf} oldap:hasRole {role.toRdf}>> oldap:hasDefaultDataPermission {dperm.toRdf} .\n'
                    rdfstar = True
            if changed:  # add the RDF*Star triples of the roles that have changed
                tmp = [role for role, dperm in changed.items() if dperm['new']]
                if tmp:
                    for role, changes in changed.items():
                        if changes['new']:
                            sparql += f'{blank:{(indent + 1) * indent_inc}}<<{self.userIri.toRdf} oldap:hasRole {role.toRdf}>> oldap:hasDefaultDataPermission {changes["new"].toRdf} .\n'
                    rdfstar = True
            if rdfstar:
                sparql += f'{blank:{(indent + 1) * indent_inc}}}}\n'
                sparql += f'{blank:{indent * indent_inc}}}}\n'
            if sparql:
                sparql_list.append(sparql)

            #
            # check if existing :Roles's have been given!
            #
            if added:
                tmp = [x.toRdf for x in added]
                tmp = ", ".join(tmp)
                ptest = f"""
                SELECT ?role
                FROM oldap:admin
                WHERE {{
                    ?role a oldap:Role .
                    FILTER(?role IN ({tmp}))
                }}
                """
                ptest_len = len(added)

        if UserAttr.IN_PROJECT in self._changeset:
            addedprojs = {}
            deletedprojs = {}
            if self._changeset[UserAttr.IN_PROJECT].action == Action.CREATE:
                addedprojs = {key: val for key, val in self._attributes[UserAttr.IN_PROJECT]}
            elif self._changeset[UserAttr.IN_PROJECT].action == Action.DELETE:
                deletedprojs = {key: val for key, val in self._changeset[UserAttr.IN_PROJECT].old_value}
            elif self._changeset[UserAttr.IN_PROJECT].action == Action.REPLACE:
                deletedprojs = {key: val for key, val in self._changeset[UserAttr.IN_PROJECT].old_value}
                addedprojs = {key: val for key, val in self._attributes[UserAttr.IN_PROJECT]}
            elif self._changeset[UserAttr.IN_PROJECT].action == Action.MODIFY:
                mod_cs = self._attributes[UserAttr.IN_PROJECT].changeset
                for key in mod_cs:
                    if mod_cs[key].action == Action.CREATE:
                        addedprojs[key] = self._attributes[UserAttr.IN_PROJECT][key]
                    elif mod_cs[key].action == Action.DELETE:
                        deletedprojs[key] = mod_cs[key].old_value

            # add projects
            if addedprojs:
                projs = [x.toRdf for x in addedprojs]
                projslist = ", ".join(projs)
                proj_test = f"""
                SELECT ?project
                FROM oldap:admin
                WHERE {{
                    ?project a oldap:Project .
                    FILTER(?project IN ({projslist}))
                }}
                """
                proj_test_len = len(addedprojs)

                sparql = f"{blank:{indent * indent_inc}}INSERT DATA {{\n"
                sparql += f'{blank:{(indent + 1) * indent_inc}}GRAPH oldap:admin {{\n'
                for proj in addedprojs:
                    sparql += f'{blank:{(indent + 2) * indent_inc}}{self.userIri.toRdf} oldap:inProject {proj.toRdf} .\n'
                    for perm in self._attributes[UserAttr.IN_PROJECT][proj]:
                        sparql += f'{blank:{(indent + 2) * indent_inc}}<<{self.userIri.toRdf} oldap:inProject {proj.toRdf}>> oldap:hasAdminPermission {perm.value} .\n'
                sparql += f'{blank:{(indent + 1) * indent_inc}}}}\n'
                sparql += f'{blank:{indent * indent_inc}}}}\n'
                sparql_list.append(sparql)

            # delete projects
            if deletedprojs:
                sparql = f"{blank:{indent * indent_inc}}DELETE DATA {{\n"
                sparql += f'{blank:{(indent + 1) * indent_inc}}GRAPH oldap:admin {{\n'
                for proj in deletedprojs:
                    sparql += f'{blank:{(indent + 2) * indent_inc}}{self.userIri.toRdf} oldap:inProject {proj.toRdf} .\n'
                    #for perm in self._changeset[UserAttr.IN_PROJECT].old_value[proj]:
                    for perm in deletedprojs[proj]:
                        sparql += f'{blank:{(indent + 2) * indent_inc}}<<{self.userIri.toRdf} oldap:inProject {proj.toRdf}>> oldap:hasAdminPermission {perm.value} .\n'
                sparql += f'{blank:{(indent + 1) * indent_inc}}}}\n'
                sparql += f'{blank:{indent * indent_inc}}}}\n'
                sparql_list.append(sparql)

            if self._changeset[UserAttr.IN_PROJECT].action == Action.MODIFY:
                in_project_cs = self._attributes[UserAttr.IN_PROJECT].changeset
                adding = {}
                deleting = {}
                delete_completely = set()
                for proj_iri in in_project_cs.keys():
                    if in_project_cs[proj_iri].action == Action.CREATE:
                        adding[proj_iri] = self._attributes[UserAttr.IN_PROJECT][proj_iri]
                    elif in_project_cs[proj_iri].action == Action.DELETE:
                        #deleting[proj_iri] = self._attributes[UserAttr.IN_PROJECT][proj_iri].old_value
                        # we have to completely delete the reference to the project!
                        deleting[proj_iri] = in_project_cs[proj_iri].old_value
                        delete_completely.add(proj_iri)
                    elif in_project_cs[proj_iri].action == Action.REPLACE:
                        adding[proj_iri] = self._attributes[UserAttr.IN_PROJECT].get(proj_iri) or set()
                        deleting[proj_iri] = in_project_cs[proj_iri].old_value
                        # if self._attributes[UserAttr.IN_PROJECT].get(proj_iri):
                        #     deleting[proj_iri] = self._attributes[UserAttr.IN_PROJECT][proj_iri].old_value
                        # else:
                        #     deleting[proj_iri] = in_project_cs[proj_iri].old_value
                    elif in_project_cs[proj_iri].action == Action.MODIFY:
                        A = self._attributes[UserAttr.IN_PROJECT][proj_iri] if self._attributes[UserAttr.IN_PROJECT].get(proj_iri) else ObservableSet()
                        B = self._attributes[UserAttr.IN_PROJECT][proj_iri].old_value if self._attributes[UserAttr.IN_PROJECT][proj_iri].old_value else ObservableSet()
                        add_set = A - B
                        del_set = B - A
                        if add_set:
                            adding[proj_iri] = add_set
                        if del_set:
                            deleting[proj_iri] = B - A

                if adding:
                    projs = [x.toRdf for x in adding.keys()]
                    projslist = ", ".join(projs)
                    proj_test = f"""
                    SELECT ?project
                    FROM oldap:admin
                    WHERE {{
                        ?project a oldap:Project .
                        FILTER(?project IN ({projslist}))
                    }}
                    """
                    proj_test_len = len(adding.keys())

                    sparql = f"{blank:{indent * indent_inc}}INSERT DATA {{\n"
                    sparql += f'{blank:{(indent + 1) * indent_inc}}GRAPH oldap:admin {{\n'
                    for proj, add_set in adding.items():
                        #perms = self._attributes[UserAttr.IN_PROJECT][proj] - self._changeset[UserAttr.IN_PROJECT].old_value[proj]
                        for perm in add_set:
                            sparql += f'{blank:{(indent + 2) * indent_inc}}<<{self.userIri.toRdf} oldap:inProject {proj.toRdf}>> oldap:hasAdminPermission {perm.value} .\n'
                    sparql += f'{blank:{(indent + 1) * indent_inc}}}}\n'
                    sparql += f'{blank:{indent * indent_inc}}}}\n'
                    sparql_list.append(sparql)
                if deleting:
                    sparql = f"{blank:{indent * indent_inc}}DELETE DATA {{\n"
                    sparql += f'{blank:{(indent + 1) * indent_inc}}GRAPH oldap:admin {{\n'
                    for proj, del_set in deleting.items():
                        #perms = self._changeset[UserAttr.IN_PROJECT].old_value[proj] - self._attributes[UserAttr.IN_PROJECT][proj]
                        for perm in del_set:
                            sparql += f'{blank:{(indent + 2) * indent_inc}}<<{self.userIri.toRdf} oldap:inProject {proj.toRdf}>> oldap:hasAdminPermission {perm.value} .\n'
                    sparql += f'{blank:{(indent + 1) * indent_inc}}}}\n'
                    sparql += f'{blank:{indent * indent_inc}}}}\n'
                    sparql_list.append(sparql)

                if delete_completely:
                    for proj_iri in delete_completely:
                        sparql = f"{blank:{indent * indent_inc}}DELETE DATA {{\n"
                        sparql += f'{blank:{(indent + 1) * indent_inc}}GRAPH oldap:admin {{\n'
                        sparql += f'{blank:{(indent + 2) * indent_inc}}{self.userIri.toRdf} oldap:inProject {proj_iri.toRdf} .\n'
                        sparql += f'{blank:{(indent + 1) * indent_inc}}}}\n'
                        sparql += f'{blank:{indent * indent_inc}}}}\n'
                        sparql_list.append(sparql)

        sparql = context.sparql_context
        sparql += " ;\n".join(sparql_list)
        self._con.transaction_start()
        try:
            modtime = self.get_modified_by_iri(Xsd_QName('oldap:admin'), self.userIri)
        except OldapError:
            self._con.transaction_abort()
            raise
        if modtime != self.modified:
            self._con.transaction_abort()
            raise OldapErrorUpdateFailed(
                f'Modifying user "{self.userId}" failed because of changed modification time: {modtime}')
        if ptest and ptest_len > 0:
            ptest_sparql = context.sparql_context
            ptest_sparql += ptest
            jsonobj = self._con.transaction_query(ptest_sparql)
            res = QueryProcessor(context, jsonobj)
            if len(res) != ptest_len:
                self._con.transaction_abort()
                raise OldapErrorValue("One of the permission sets is not existing!")

        if proj_test and proj_test_len > 0:
            proj_sparql = context.sparql_context
            proj_sparql += proj_test
            jsonobj = self._con.transaction_query(proj_sparql)
            res = QueryProcessor(context, jsonobj)
            if len(res) != proj_test_len:
                self._con.transaction_abort()
                raise OldapErrorValue("One of the projects is not existing!")

        try:
            self._con.transaction_update(sparql)
            self.set_modified_by_iri(Xsd_QName('oldap:admin'), self.userIri, self.modified, timestamp)
            modtime = self.get_modified_by_iri(Xsd_QName('oldap:admin'), self.userIri)
        except OldapError:
            print(sparql)
            self._con.transaction_abort()
            raise
        if timestamp != modtime:
            raise OldapErrorUpdateFailed(f"Update failed! Timestamp does not match (timestamp: {timestamp}, modtime: {modtime}).")
        try:
            self._con.transaction_commit()
        except OldapError:
            self._con.transaction_abort()
            raise
        self._modified = timestamp
        self._contributor = self._con.userIri
        cache = CacheSingletonRedis()
        cache.set(self.userIri, self)




import json
from copy import deepcopy
from datetime import datetime
from enum import Enum
from pprint import pprint
from typing import Set, Dict, Any, Self

from oldaplib.src.connection import Connection
from oldaplib.src.enums.action import Action
from oldaplib.src.enums.attributeclass import AttributeClass
from oldaplib.src.enums.language import Language
from oldaplib.src.helpers.attributechange import AttributeChange
from oldaplib.src.helpers.context import Context
from oldaplib.src.helpers.numeric import Numeric
from oldaplib.src.helpers.serializer import serializer
from oldaplib.src.xsd.iri import Iri
from oldaplib.src.xsd.xsd_boolean import Xsd_boolean
from oldaplib.src.xsd.xsd_qname import Xsd_QName
from oldaplib.src.xsd.xsd_datetime import Xsd_dateTime
from oldaplib.src.helpers.oldaperror import OldapError, OldapErrorNotFound, OldapErrorType, OldapErrorImmutable, OldapErrorValue
from oldaplib.src.helpers.query_processor import QueryProcessor
from oldaplib.src.iconnection import IConnection


#@strict
@serializer
class Model:
    """
    Model class manages attributes, encapsulates state, and tracks changes.

    This class is intended to provide a concrete implementation for objects with
    specific attributes, ensuring consistency, change tracking, and optional
    validation logic. The primary purpose of this class is to be extended or used
    as a foundational class in a larger system. It supports deep copying,
    notification mechanisms, and attribute manipulation while maintaining an
    immutable core structure for sensitive attributes.

    :ivar creator: The creator of the project, represented as an IRI.
    :type creator: Iri | None
    :ivar created: The creation date of the project, represented as a datetime object.
    :type created: Xsd_dateTime | None
    """
    _con: IConnection
    _changed: Set[str]
    _creator: Iri | None
    _created: Xsd_dateTime | None
    _contributor: Iri | None
    _modified: Xsd_dateTime | None
    _attributes: dict[AttributeClass, Any]
    _changeset: dict[AttributeClass | Xsd_QName, AttributeChange]
    _validate: bool

    def __init__(self, *,
                 connection: IConnection,
                 creator: Iri | str | None = None,
                 created: Xsd_dateTime | datetime | str | None = None,
                 contributor: Iri | str | None = None,
                 modified: Xsd_dateTime | datetime | str | None = None,
                 validate: bool = False) -> None:
        """
        Initializes an instance of the class.

        :param connection: An instance of `IConnection` representing the current
            connection.
        :param creator: An optional parameter representing the creator, which can
            be a string, an IRI object, or None.
        :param created: An optional parameter representing the creation date/time,
            which can be an `Xsd_dateTime`, `datetime`, `str`, or None.
        :param contributor: An optional parameter representing the contributor,
            which can be a string, an IRI object, or None.
        :param modified: An optional parameter representing the last modification
            date/time, which can be an `Xsd_dateTime`, `datetime`, `str`, or None.
        :param validate: Boolean specifying whether to validate input values.

        :raises OldapError: If the `connection` parameter is not an instance of
            `IConnection`.
        :raises OldapErrorValue: If the `creator`, `created`, or `contributor`
            parameters are not valid IRI objects.
        """
        if not isinstance(connection, IConnection):
            raise OldapError('"connection"-parameter must be an instance of IConnection')
        self._validate = validate
        self._con = connection
        if not creator:
            creator = self._con.userIri
        if not contributor:
            contributor = self._con.userIri

        self._creator = Iri(creator, validate=validate)
        self._created = Xsd_dateTime(created, validate=validate)
        self._contributor = Iri(contributor, validate=validate)
        self._modified = Xsd_dateTime(modified, validate=validate)
        self._attributes = {}
        self._changeset = {}

    def _as_dict(self) -> dict[str, Any]:
        return {
            **({"creator": self._creator} if self._creator is not None else {}),
            **({"created": self._created} if self._created is not None else {}),
            **({"contributor": self._contributor} if self._contributor is not None else {}),
            **({"modified": self._modified} if self._modified is not None else {}),
        }

    def __deepcopy__(self, memo: Dict[int, Any]) -> 'Model':
        cls = self.__class__
        instance = cls.__new__(cls)
        memo[id(self)] = instance
        instance._creator = deepcopy(self._creator, memo=memo)
        instance._created = deepcopy(self._created, memo=memo)
        instance._contributor = deepcopy(self._contributor, memo=memo)
        instance._modified = deepcopy(self._modified, memo=memo)
        instance._attributes = deepcopy(self._attributes, memo=memo)
        instance._changeset = {}

        return instance

    def check_consistency(self, attr: AttributeClass, value: Any) -> None:
        pass

    def pre_transform(self, attr: AttributeClass, value: Any, validate: bool = False) -> Any:
        return value

    def cleanup_setter(self, attr: AttributeClass, value: Any) -> None:
        pass

    def notifier(self, attr: AttributeClass, value: Any) -> None:
        pass

    def __str__(self) -> str:
        res = f'Creation: {self._created} by {self._creator}\n'
        res += f'Modified: {self._modified} by {self._contributor}\n'
        for attr, value in self._attributes.items():
            res += f'{attr} ({attr.value}): {value}\n'
        return res

    def __getitem__(self, attr: AttributeClass) -> Any:
        return self._attributes.get(attr)

    def get(self, attr: AttributeClass) -> Any:
        return self._attributes.get(attr)

    def __setitem__(self, attr: AttributeClass, value: Any) -> None:
        self._change_setter(attr, value)

    def __delitem__(self, attr: AttributeClass) -> None:
        if self._attributes.get(attr) is not None:
            self._changeset[attr] = AttributeChange(self._attributes[attr], Action.DELETE)
            del self._attributes[attr]

    # def __delattr__(self, attrstr: str) -> None:
    #     try:
    #         attr = AttributeClass.from_name(attrstr)
    #     except ValueError:
    #         raise OldapErrorValue(f'Nonexisting attribute: "{attrstr}"')
    #     if self._attributes.get(attr) is not None:
    #         self._changeset[attr] = AttributeChange(self._attributes[attr], Action.DELETE)
    #         del self._attributes[attr]

    def set_attributes(self, arguments: dict[str, Any], Attributes: type[Enum]) -> None:
        """
        Sets attributes on the instance using the provided arguments and attributes metadata.

        :param arguments: A dictionary mapping attribute names to their values.
        :type arguments: dict[str, Any]
        :param Attributes: An enumeration class representing the possible attributes, their types,
            and validation metadata.
        :type Attributes: type[Enum]
        :raises OldapErrorValue: If a provided value cannot be converted to the specific attribute's datatype.
        :raises OldapErrorType: If a mandatory attribute is missing in the given arguments.
        :return: None
        """
        for name, value in arguments.items():
            if not isinstance(value, (bool, Xsd_boolean)) and not value:
                continue
            attr = Attributes.from_name(name)
            try:
                self._attributes[attr] = value if isinstance(value, attr.datatype) else attr.datatype(value, validate=self._validate)
            except ValueError as err:
                raise OldapErrorValue(err)
            if hasattr(self._attributes[attr], 'set_notifier'):
                self._attributes[attr].set_notifier(self.notifier, attr)
        for attr in Attributes:
            if attr.mandatory and not self._attributes.get(attr):
                raise OldapErrorType(f'Mandatory parameter {attr.name} is missing.')

    def _get_value(self: Self, attr: AttributeClass) -> Any | None:
        tmp = self._attributes.get(attr)
        if attr.datatype != Xsd_boolean and not tmp:
            return None
        return tmp

    def _set_value(self: Self, value: Any, attr: AttributeClass) -> None:
        self._change_setter(attr, value)

    def _del_value(self: Self, attr: AttributeClass) -> None:
        self._changeset[attr] = AttributeChange(self._attributes[attr], Action.DELETE)
        del self._attributes[attr]
        if hasattr(self, "notify"):
            self.notify()
        self.cleanup_setter(attr, None)

    def _change_setter(self, attr: AttributeClass, value: Any) -> None:
        if value is not None:
            if self._attributes.get(attr) == value:
                return
            if attr.immutable:
                raise OldapErrorImmutable(f'Attribute {attr.value} is immutable.')
            self.check_consistency(attr, value)
            value = self.pre_transform(attr, value, validate=self._validate)
        if self._attributes.get(attr) is None:
            if self._changeset.get(attr) is None:
                self._changeset[attr] = AttributeChange(None, Action.CREATE)
        else:
            if value is None:
                if self._changeset.get(attr) is None:
                    self._changeset[attr] = AttributeChange(deepcopy(self._attributes[attr]), Action.DELETE)
            else:
                if self._changeset.get(attr) is None:
                    self._changeset[attr] = AttributeChange(deepcopy(self._attributes[attr]), Action.REPLACE)
        if value is None:
            if self._attributes.get(attr) is not None:
                del self._attributes[attr]
        else:
            if not isinstance(value, attr.datatype):
                try:
                    self._attributes[attr] = attr.datatype(value, validate=True)
                except Exception as err:
                    raise Exception(f"Failed to set attribute '{attr}' with value '{value}': {err}")
            else:
                self._attributes[attr] = value
            if hasattr(self._attributes[attr], 'set_notifier') and hasattr(self, 'notifier'):
                self._attributes[attr].set_notifier(self.notifier, attr)
        if hasattr(self, "notify"):
            self.notify()
        self.cleanup_setter(attr, value)

    @property
    def changeset(self) -> Dict[AttributeClass, AttributeChange]:
        """
        Provides access to the collection of changes, represented as a
        dictionary, where keys are attributes and values are the corresponding
        changes. This property is useful for internal use or debugging purposes.

        :return: A dictionary containing the mapping of attributes to their
            respective changes.
        :rtype: Dict[AttributeClass, AttributeChange]
        """
        return self._changeset

    def clear_changeset(self) -> None:
        """
        Clear the changeset for all applicable attributes. This method is utilized
        to reset tracked changes, an internal operation primarily used for debugging
        or restoration of a clean state in the managed object or its nested properties.

        :return: None
        """
        for item in self._attributes:
            if hasattr(self._attributes[item], 'clear_changeset'):
                self._attributes[item].clear_changeset()
        self._changeset = {}

    @property
    def creator(self) -> Iri | None:
        """
        The creator of the project.
        :return: Iri of the creator of the project.
        :rtype: Iri | None
        """
        return self._creator

    @property
    def created(self) -> Xsd_dateTime | None:
        """
        The creation date of the project.
        :return: Creation date of the project.
        :rtype: Xsd_dateTime | None
        """
        return self._created

    @property
    def contributor(self) -> Iri | None:
        """
        The contributor of the project as Iri.
        :return: Iri of the contributor of the project.
        :rtype: Iri | None
        """
        return self._contributor

    @property
    def modified(self) -> Xsd_dateTime | None:
        """
        Modification date of the project.
        :return: Modification date of the project.
        :rtype: Xsd_dateTime | None
        """
        return self._modified


    def get_modified_by_iri(self, graph: Xsd_QName | str, iri: Iri | str) -> Xsd_dateTime:
        """
        This method retrieves the modification timestamp of a resource identified by its IRI
        from a specified RDF graph. It executes a SPARQL query to extract the required
        information. The method ensures type validation for the input parameters and
        follows appropriate context operations for querying.

        :param graph: The target RDF graph from which the modification data will be retrieved.
                      This can be either an `Xsd_QName` object or a `str` representing the graph name.
        :param iri: The IRI of the resource to query for its modification timestamp. It can
                    be provided as an `Iri` object or a `str`.
        :return: The modification timestamp of the requested resource as an `Xsd_dateTime` object.
        :raises OldapErrorNotFound: If no resource is found with the specified IRI in the graph.
        """
        if not isinstance(graph, Xsd_QName):
            graph = Xsd_QName(graph, validate=True)
        if not isinstance(iri, Iri):
            iri = Iri(iri, validate=True)
        context = Context(name=self._con.context_name)
        sparql = context.sparql_context
        sparql += f"""
        SELECT ?modified
        FROM {graph}
        WHERE {{
            {iri.toRdf} dcterms:modified ?modified
        }}
        """
        jsonobj = None
        if self._con.in_transaction():
            jsonobj = self._con.transaction_query(sparql)
        else:
            jsonobj = self._con.query(sparql)
        res = QueryProcessor(context, jsonobj)
        if len(res) != 1:
            raise OldapErrorNotFound(f'No resource found with iri "{iri}".')
        for r in res:
            return r['modified']

    def set_modified_by_iri(self,
                            graph: Xsd_QName | str,
                            iri: Iri | str,
                            old_timestamp: Xsd_dateTime | datetime | str,
                            timestamp: Xsd_dateTime | datetime | str) -> None:
        """
        Updates the modification timestamp and contributor information associated
        with the given IRI in a specified graph. This operation involves querying
        with a SPARQL DELETE/INSERT statement and substituting the old timestamp
        with the new one while associating the resource with the current user as
        the contributor.

        This method ensures that the parameters are instantiated as appropriate
        types (`Xsd_QName`, `Iri`, `Xsd_dateTime`) if they are not already so.

        :param graph: The graph in which the modification and contributor information
                      needs to be updated. Can be provided as an instance of `Xsd_QName`
                      or a string.
        :param iri: The IRI of the resource whose modification information is being
                    updated. Can be provided as an instance of `Iri` or a string.
        :param old_timestamp: The existing timestamp to be removed. Can be provided as
                              an instance of `Xsd_dateTime`, `datetime`, or a string.
        :param timestamp: The new timestamp to be set. Can be provided as an instance
                          of `Xsd_dateTime`, `datetime`, or a string.
        :return: None
        :raises OldapError: If an exception occurs during the operation, it is wrapped
                            in an `OldapError` and raised.
        """
        if not isinstance(graph, Xsd_QName):
            graph = Xsd_QName(graph, validate=True)
        if not isinstance(iri, Iri):
            iri = Iri(iri, validate=True)
        if not isinstance(old_timestamp, Xsd_dateTime):
            old_timestamp = Xsd_dateTime(old_timestamp, validate=True)
        if not isinstance(timestamp, Xsd_dateTime):
            timestamp = Xsd_dateTime(timestamp, validate=True)
        try:
            context = Context(name=self._con.context_name)
            sparql = context.sparql_context
            sparql += f"""
            WITH {graph}
            DELETE {{
                ?res dcterms:modified {old_timestamp.toRdf} .
                ?res dcterms:contributor ?contributor .
            }}
            INSERT {{
                ?res dcterms:modified {timestamp.toRdf} .
                ?res dcterms:contributor {self._con.userIri.toRdf} .
            }}
            WHERE {{
                BIND({iri.toRdf} as ?res)
                ?res dcterms:modified {old_timestamp.toRdf} .
                ?res dcterms:contributor ?contributor .
            }}
            """
        except Exception as e:
            raise OldapError(e)
        if self._con.in_transaction():
            self._con.transaction_update(sparql)
        else:
            self._con.update_query(sparql)

    def safe_query(self, query: str) -> Any:
        """
        Executes a safe transaction query and ensures any potential errors during
        execution are handled appropriately. If an error occurs, the transaction
        is aborted to maintain consistency.

        :param query: The SQL query string to execute within the context of a
                      database transaction.
        :type query: str

        :return: The result of the query execution, of any type, depending on the
                 database operation performed.
        :rtype: Any

        :raises OldapError: If an error occurs during the query execution. The
                            transaction is aborted before re-raising the error.
        """
        try:
            return self._con.transaction_query(query)
        except OldapError:
            self._con.transaction_abort()
            raise

    def safe_update(self, update_query: str) -> None:
        """
        Executes a safe update operation within a transaction context. If an error
        occurs during the update, the transaction is aborted.

        :param update_query: The SQL update query to be executed.
        :type update_query: str
        :return: None
        :rtype: None
        :raises OldapError: If the update operation fails, the error is propagated
            after the transaction is rolled back.
        """
        try:
            self._con.transaction_update(update_query)
        except OldapError:
            self._con.transaction_abort()
            raise

    def safe_commit(self) -> None:
        """
        Commits a transaction safely. If the commit operation fails due to an OldapError,
        the transaction is aborted to ensure data consistency.

        :return: None
        :raises OldapError: If the transaction commit fails and it cannot be
            successfully completed.
        """
        try:
            self._con.transaction_commit()
        except OldapError:
            self._con.transaction_abort()
            raise




"""
# XSD Datatypes, IRI Validator, XSD Validator
"""
import re
from urllib.parse import urlparse

import xmlschema
from enum import Enum, unique
from pystrict import strict
from typing import Any

from oldaplib.src.helpers.serializer import serializer


@unique
@serializer
class XsdDatatypes(Enum):
    """
    Enumeration of the supported xsd datatypes:

    - `XsdDatatypes.string` = 'xsd:string'
    - `XsdDatatypes.langString` = 'rdf:langString'
    - `XsdDatatypes.boolean` = 'xsd:boolean'
    - `XsdDatatypes.decimal` = 'xsd:decimal'
    - `XsdDatatypes.float` = 'xsd:float'
    - `XsdDatatypes.double` = 'xsd:double'
    - `XsdDatatypes.duration` = 'xsd:duration'
    - `XsdDatatypes.dateTime` = 'xsd:dateTime'
    - `XsdDatatypes.dateTimeStamp` = 'xsd:dateTimeStamp'
    - `XsdDatatypes.time` = 'xsd:time'
    - `XsdDatatypes.date` = 'xsd:date'
    - `XsdDatatypes.gYearMonth` = 'xsd:gYearMonth'
    - `XsdDatatypes.gYear` = 'xsd:gYear'
    - `XsdDatatypes.gMonthDay` = 'xsd:gMonthDay'
    - `XsdDatatypes.gDay` = 'xsd:gDay'
    - `XsdDatatypes.gMonth` = 'xsd:gMonth'
    - `XsdDatatypes.hexBinary` = 'xsd:hexBinary'
    - `XsdDatatypes.base64Binary` = 'xsd:base64Binary'
    - `XsdDatatypes.anyURI` = 'xsd:anyURI'
    - `XsdDatatypes.QName` = 'xsd:QName'
    - `XsdDatatypes.normalizedString` = 'xsd:normalizedString'
    - `XsdDatatypes.token` = 'xsd:token'
    - `XsdDatatypes.language` = 'xsd:language'
    - `XsdDatatypes.Name` = 'xsd:name'
    - `XsdDatatypes.NCName` = 'xsd:NCName'
    - `XsdDatatypes.NMTOKEN` = 'xsd:NMTOKEN'
    - `XsdDatatypes.ID` = 'xsd:ID'
    - `XsdDatatypes.IDREF` = 'xsd:IDREF'
    - `XsdDatatypes.IDREFS` = 'xsd: IDREFS'
    - `XsdDatatypes.integer` = 'xsd:integer'
    - `XsdDatatypes.nonPositiveInteger` = 'xsd:nonPositiveInteger'
    - `XsdDatatypes.unsignedLong` = 'xsd:unsignedLong'
    - `XsdDatatypes.unsignedInt` = 'xsd:unsignedInt'
    - `XsdDatatypes.unsignedShort` = 'xsd:unsignedShort'
    - `XsdDatatypes.unsignedByte` = 'xsd:unsignedByte'
    - `XsdDatatypes.positiveInteger` = 'xsd:positiveInteger'

    """
    string = 'xsd:string'
    langString = 'rdf:langString'
    boolean = 'xsd:boolean'
    decimal = 'xsd:decimal'
    float = 'xsd:float'
    double = 'xsd:double'
    duration = 'xsd:duration'
    dateTime = 'xsd:dateTime'
    dateTimeStamp = 'xsd:dateTimeStamp'
    time = 'xsd:time'
    date = 'xsd:date'
    gYearMonth = 'xsd:gYearMonth'
    gYear = 'xsd:gYear'
    gMonthDay = 'xsd:gMonthDay'
    gDay = 'xsd:gDay'
    gMonth = 'xsd:gMonth'
    hexBinary = 'xsd:hexBinary'
    base64Binary = 'xsd:base64Binary'
    anyURI = 'xsd:anyURI'
    QName = 'xsd:QName'
    normalizedString = 'xsd:normalizedString'
    token = 'xsd:token'
    language = 'xsd:language'
    name_ = 'xsd:name'
    NCName = 'xsd:NCName'
    NMTOKEN = 'xsd:NMTOKEN'
    ID = 'xsd:ID'
    IDREF = 'xsd:IDREF'
    integer = 'xsd:integer'
    nonPositiveInteger = 'xsd:nonPositiveInteger'
    negativeInteger = 'xsd:negativeInteger'
    long = 'xsd:long'
    int = 'xsd:int'
    short = 'xsd:short'
    byte = 'xsd:byte'
    nonNegativeInteger = 'xsd:nonNegativeInteger'
    unsignedLong = 'xsd:unsignedLong'
    unsignedInt = 'xsd:unsignedInt'
    unsignedShort = 'xsd:unsignedShort'
    unsignedByte = 'xsd:unsignedByte'
    positiveInteger = 'xsd:positiveInteger'

    def __str__(self):
        return self.value

    @property
    def toRdf(self):
        return self.value


@strict
class IriValidator:
    """
    Calls to validate the syntax of an IRI
    """
    @classmethod
    def validate(cls, val: str) -> bool:
        """
        Class method which validates an IRI. Supported are the protocols "http", "https", and "urn"

        :param val: String to be validated as IRI
        :type val: str
        :return: True or False
        """
        try:
            result = urlparse(val)
            match result.scheme:
                case 'http':
                    return True if result.netloc else False
                case 'https':
                    return True if result.netloc else False
                case 'urn':
                    return True if result.path else False
                case _:
                    return False
            #return all([result.scheme, result.netloc])
        except Exception:
            return False

@strict
class XsdValidator:
    """
    Class to validate generic XSD datatypes
    """
    @classmethod
    def validate(cls,
                 datatype: XsdDatatypes,
                 value: Any) -> bool:
        """
        Validate a value against the given xsd datatype. It uses the xmlschema.XMLSchema11 validator.

        :param datatype: The xsd datatype the value should be validated against
        :type datatype: XsdDatatypes
        :param value: A value to be validated
        :return: True or False
        """
        xsd_string = f"""
        <xsd:schema xmlns:xsd="http://www.w3.org/2001/XMLSchema">
            <xsd:element name="tag" type="{datatype.value}"/>
        </xsd:schema>
        """
        namespace = ""
        if datatype == XsdDatatypes.QName:
            parts = str(value).split(':')
            if len(parts) == 2:
                namespace = f"xmlns:{parts[0]}=\"http://dummy.net/dum\""
            else:
                namespace = "xmlns:dummy=\"http://dummy.net/dum\""
        elif datatype == XsdDatatypes.anyURI:
            return IriValidator.validate(value)
        try:
            xsd_validator = xmlschema.XMLSchema11(xsd_string)
            return xsd_validator.is_valid(f"""<?xml version="1.0" encoding="UTF-8"?><tag {namespace}>""" + str(value) + "</tag>")
        except xmlschema.XMLSchemaParseError as e:
            return False


if __name__ == '__main__':
    print(XsdValidator.validate(XsdDatatypes.anyURI, "http://waelo.org/data/gaga#uuu"))
    print(XsdValidator.validate(XsdDatatypes.anyURI, "https://waelo.org/data/gaga#uuu"))
    print(XsdValidator.validate(XsdDatatypes.anyURI, "https://waelo.org/data/gaga#"))
    print(XsdValidator.validate(XsdDatatypes.NMTOKEN, "https://orcid.org/0000-0003-1681-4036"))
    print(XsdValidator.validate(XsdDatatypes.anyURI, "urn:uuid:7e56b6c4-42e5-4a9d-94cf-d6e22577fb4b"))
    print(XsdValidator.validate(XsdDatatypes.anyURI, "waselias"))
